unit lexout;

{$MODE DELPHI}{$H+}

interface

uses classes, sysutils, lexin, relant, splex;

const
  C_LB_MSG_BUILD_NFA = ' > Building NFA for %s ';
  C_LB_MSG_BUILD_UNFA = ' >>> Building NFA for Token %s from RegEx (%s)';
  C_LB_MSG_BUILD_DFA = ' > Building DFA for %s ';
  C_LB_MSG_BUILD_MinDFA = ' > Building MinDFA for %s ';

  C_LB_MSG_VER_NFA_FAILED = ' - Verification for NFA %s failed';
  C_LB_MSG_VER_DFA_FAILED = ' - Verification for DFA %s failed ';

  C_LB_MSG_STATS_NFA = '   NFA Stats: %d Nodes, %d Transitions ';
  C_LB_MSG_STATS_DFA = '   DFA Stats: %d Nodes, %d Transitions ';

  C_LB_IE_BUILD_NFA = ' * Building NFA raised Exception: %s ';
  C_LB_IE_BUILD_DFA = ' * Building DFA raised Exception: %s ';
  C_LB_IE_BUILD_MinDFA = ' * Building MinDFA raised Exception: %s ';

type
  { TLexerBuilder }

  TLexerBuilder = class
    private
      FLexerDefinition: TLexerDefinition;
      FNFA: TNFA;
      FVStream: TStrings;
      FEStream: TStrings;
      FOStream: TStrings;
      FVerbose: Boolean;
      FInternalError: Boolean;
      FStats: Boolean;
    protected
      function constructNFA: Boolean;
      function constructDFA: Boolean;
      function constructMinDFA: Boolean;
      procedure StatsOut(const msg: String);
      procedure VerboseOut( const msg: String );
      procedure IErrorOut( const msg: String );
      procedure WriteO( const s: String );
      procedure WriteLnO( const s: String );
    public
      property VStream: TStrings read FVStream write FVStream;
      property EStream: TStrings read FEStream write FEStream;
      property OStream: TStrings read FOStream write FOStream;
      property Verbose: Boolean read FVerbose write FVerbose;
      property Stats: Boolean read FStats write FStats;
      property InternalError: Boolean read FInternalError write FInternalError;
      property LexerDefinition: TLexerDefinition read FLexerDefinition;
      property CurrentNFA: TNFA read FNFA;
      procedure Generate; virtual;
      constructor Create( ldef: TLexerDefinition ); virtual;
      destructor Destroy; override;
  end;

  TNFADotBuilderType = ( dbtNFA, dbtDFA, dbtMinDFA );
  TNFADotBuilderTypeSet = set of TNFADotBuilderType;

  { TNFADotBuilder }

  TNFADotBuilder = class( TLexerBuilder )
    private
      FOutTypes: TNFADotBuilderTypeSet;
      procedure generateCluster( const l: String );
      function generateNFA: Boolean;
      function generateDFA: Boolean;
      function generateMinDFA: Boolean;
    public
      property OutputTypes: TNFADotBuilderTypeSet read FOutTypes write FOutTypes;
      constructor Create( ldef: TLexerDefinition ); override;
      procedure Generate; override;
  end;

  { TSPLexDFABuilder }

  TSPLexDFABuilder = class( TLexerBuilder )
    protected
      FAlphas: array of TAlphaSet;
      function getAlphaRef( const a: TAlphaSet ): Integer;
    public
      constructor Create( ldef: TLexerDefinition ); override;
      procedure Generate; override;
  end;

  { TSPLexDynamicScanDFA }

  TSPLexDynamicScanDFA = class( TSPLexScanDFA )
    protected
      FFinals: array of String;
      procedure SetFinals( i: Integer );
      procedure SetFinalName( i: Integer; const s: String );
      procedure SetNodeTableLength( i: Integer );
      procedure SetNodeState( i, s: Integer );
      procedure SetNodeDeltas( i, d: Integer );
      procedure SetNodeDeltaSigma( i, d, aref: Integer );
      procedure SetNodeDeltaDest( i, d, t: Integer );
      procedure SetSigmaTableLength( i: Integer );
      procedure SetSigma( i: Integer; const s: TSPLexSigmaSet );
    public
      function GetFinalName( i: Integer ): String;
      constructor Create; override;
      destructor Destroy; override;
  end;

  { TSPLexDynamicDFABuilder }

  TSPLexDynamicDFABuilder = class( TSPLexDFABuilder )
    private
      FSPLexDFA: TSPLexDynamicScanDFA;
    public
      property SPLexDynDFA: TSPLexDynamicScanDFA read FSPLexDFA write FSPLexDFA;
      constructor Create( ldef: TLexerDefinition ); override;
      procedure Generate; override;
      destructor Destroy; override;
  end;

implementation

{ TLexerBuilder }

function TLexerBuilder.constructNFA: Boolean;
var i: Integer;
    unfa: TUNFA;
begin
  Result := true;
  if FNFA <> nil then
  begin
    FNFA.Free;
    FNFA := nil;
  end;
  FNFA := TNFA.Create;
  VerboseOut( Format( C_LB_MSG_BUILD_NFA, [ FLexerDefinition.LexerName ] ) );
  try
    for i := 0 to High( FLexerDefinition.Tokens.RegExs ) do
    begin
      unfa := nil;
      VerboseOut( Format( C_LB_MSG_BUILD_UNFA,
        [ FLexerDefinition.Tokens.RegExs[ i ].FinalName,
          FLexerDefinition.Tokens.RegExs[ i ].ToString ] ) );
      unfa := RegExToUNFA( FLexerDefinition.Tokens.RegExs[ i ].FinalName,
                           FLexerDefinition.Tokens.RegExs[ i ] );
      AddUNFAToNFA( unfa, FNFA );
      unfa.Free;
    end;
    Result := VerifyNFA( FNFA );
    if not Result then
      VerboseOut( Format( C_LB_MSG_VER_NFA_FAILED,
        [ FLexerDefinition.Tokens.RegExs[ i ].FinalName ] ) );
    if FStats and Result then
      StatsOut( Format( C_LB_MSG_STATS_NFA,
        [ DeltaNodeCount( FNFA ), DeltaTransCount( FNFA ) ] ) );
  except on e: Exception do
    begin
      Result := false;
      IErrorOut( Format( C_LB_IE_BUILD_NFA, [ e.Message ] ) );
      if unfa <> nil then unfa.Free;
    end;
  end;
end;

function TLexerBuilder.constructDFA: Boolean;
var tmpNFA: TNFA;
begin
  Result := true;
  if ( FNFA <> nil ) and ( FNFA.ConstructedMinDFA ) then
  begin
    FNFA.Free;
    FNFA := nil;
  end;
  if FNFA = nil then Result := constructNFA;
  if Result and ( not FNFA.ConstructedDFA ) then
  begin
    VerboseOut( Format( C_LB_MSG_BUILD_DFA, [ FLexerDefinition.LexerName ] ) );
    try
      tmpNFA := BuildDFA( FNFA );
      if VerifyDFA( tmpNFA ) then
      begin
        FNFA.Free;
        FNFA := tmpNFA;
        if FStats then
          StatsOut( Format( C_LB_MSG_STATS_DFA,
            [ DeltaNodeCount( FNFA ), DeltaTransCount( FNFA ) ] ) );
      end else
      begin
        VerboseOut( Format( C_LB_MSG_VER_DFA_FAILED, [ FLexerDefinition.LexerName ] ) );
        tmpNFA.Free;
        Result := false;
      end;
    except on e: Exception do
      begin
        Result := false;
        IErrorOut( Format( C_LB_IE_BUILD_DFA, [ e.Message ] ) );
      end;
    end;
  end;
end;

function TLexerBuilder.constructMinDFA: Boolean;
var tmpNFA: TNFA;
begin
  Result := true;
  if ( FNFA = nil ) or ( not FNFA.ConstructedDFA ) then Result := constructDFA;
  if Result and ( not FNFA.ConstructedMinDFA ) then
  begin
    VerboseOut( Format( C_LB_MSG_BUILD_MinDFA, [ FLexerDefinition.LexerName ] ) );
    try
      tmpNFA := BuildMinDFA( FNFA );
      if VerifyDFA( tmpNFA ) then
      begin
        FNFA.Free;
        FNFA := tmpNFA;
        if FStats then
          StatsOut( Format( C_LB_MSG_STATS_DFA,
            [ DeltaNodeCount( FNFA ), DeltaTransCount( FNFA ) ] ) );
      end else
      begin
        VerboseOut( Format( C_LB_MSG_VER_DFA_FAILED, [ FLexerDefinition.LexerName ] ) );
        tmpNFA.Free;
        Result := false;
      end;
    except on e: Exception do
      begin
        Result := false;
        IErrorOut( Format( C_LB_IE_BUILD_MinDFA, [ e.Message ] ) );
      end;
    end;
  end;
end;

procedure TLexerBuilder.StatsOut(const msg: String);
begin
  if FStats and Assigned(FVStream) then
    FVStream.Add( msg );
end;

procedure TLexerBuilder.VerboseOut(const msg: String);
begin
  if FVerbose and Assigned(FVStream) then
    FVStream.Add( msg );
end;

procedure TLexerBuilder.IErrorOut(const msg: String);
begin
  if FInternalError and Assigned(FEStream) then
    FEStream.Add( msg );
end;

procedure TLexerBuilder.WriteO(const s: String);
begin
  if Assigned(FOStream) then FOStream.Add( s );
end;

procedure TLexerBuilder.WriteLnO(const s: String);
begin
  if Assigned(FOStream) then FOStream.Add( s );
end;

procedure TLexerBuilder.Generate;
begin
  constructMinDFA;
end;

constructor TLexerBuilder.Create(ldef: TLexerDefinition);
begin
  FLexerDefinition := ldef;
  FNFA := nil;
  FVStream := nil;
  FEStream := nil;
  FOStream := nil;
  FVerbose := false;
  FInternalError := true;
  FStats := false;
end;

destructor TLexerBuilder.Destroy;
begin
  inherited Destroy;
  If FNFA <> nil then FNFA.Free;
end;

{ TNFADotBuilder }

function AlphaSetToDotString( const s: TAlphaSet ): String;
var ais: TASCIIInputSet;
    tmps: String;
    i: Integer;
begin
  Result := '';
  ais := TASCIIInputSet.Create( TASCIIAlphaSet(s) );
  tmps := ais.AlphaSetToStringStripApo;
  ais.Free;
  i := 1;
  while i <= Length( tmps ) do
  begin
    if tmps[ i ] <> '\' then
    begin
      if tmps[ i ] <> '"' then
      begin
        Result := Result + tmps[ i ];
      end else
      begin
        Result := Result + '\"';
      end;
    end else
    begin
      Result := Result + '\\';
    end;
    Inc( i, 1 );
  end;
end;

procedure TNFADotBuilder.generateCluster(const l: String);
var i: TFANodeType;
    nodeAttr: String;
    nodeLabel: String;
    titer: TNFATransitionIterator;

  function nodeName( i: Integer ): String;
  begin
    Result := 'Node_'+l+'_'+IntToStr( i );
  end;

begin
  WriteLnO( ' subgraph cluster_'+l+' {' );
  WriteLnO( '    label = "'+l+'";' );
  // Write Nodes with FinalState
  for i := Low( CurrentNFA.Nodes ) to High( CurrentNFA.Nodes ) do
  begin
    nodeLabel := 'Node '+IntToStr( i );
    if CurrentNFA.FinalState[ i ] >= TFAFinalMin then
    begin
      nodeAttr := ', shape = doublecircle';
      nodeLabel := nodeLabel + ' (' +
        CurrentNFA.FinalNames[ CurrentNFA.FinalState[ i ] ] + ')';
    end else nodeAttr := ', shape = circle';
    WriteLnO( '  '+nodeName( i )+' [label="'+nodeLabel+'" '+nodeAttr+']' );
  end;
  WriteLnO( '' );
  // Write Transitions
  for i := Low( CurrentNFA.Nodes ) to High( CurrentNFA.Nodes ) do
  begin
    titer := CurrentNFA.Nodes[ i ].TransIterator;
    while titer.Next do
    begin
      WriteLnO( '  '+nodeName( i )+'->'+nodeName( titer.CurrentDestNode )+
                ' [label="'+AlphaSetToDotString( titer.CurrentTransSet )+'"]' );
    end;
    titer.Free;
  end;
  WriteLnO( '  }' );
end;

function TNFADotBuilder.generateNFA: Boolean;
begin
  Result := constructNFA;
  if Result then generateCluster( 'NFA' );
end;

function TNFADotBuilder.generateDFA: Boolean;
begin
  Result := constructDFA;
  if Result then generateCluster( 'DFA' );
end;

function TNFADotBuilder.generateMinDFA: Boolean;
begin
  Result := constructMinDFA;
  if Result then generateCluster( 'MinDFA' );
end;

constructor TNFADotBuilder.Create(ldef: TLexerDefinition);
begin
  inherited Create(ldef);
  FOutTypes := [ dbtNFA, dbtDFA, dbtMinDFA ];
end;

procedure TNFADotBuilder.Generate;
var noerror: Boolean;
begin
  WriteLnO( 'digraph splexer {' );
  WriteLnO( '  rankdir=LR;' );
  WriteLnO( '  NodeStart [style = invis];' );
  noerror := true;
  if dbtNFA in FOutTypes then
    noerror := GenerateNFA;
  if noerror and ( dbtDFA in FOutTypes ) then
    noerror := GenerateDFA;
  if noerror and ( dbtMinDFA in FOutTypes ) then
    noerror := GenerateMinDFA;
  WriteLnO( '}' );
end;

  { TSPLexDFABuilder }

function AlphaSetToString( const s: TAlphaSet ): String;
var ais: TASCIIInputSet;
begin
  ais := TASCIIInputSet.Create( TASCIIAlphaSet(s) );
  Result := ais.AlphaSetToString;
  ais.Free;
end;

function TSPLexDFABuilder.getAlphaRef( const a: TAlphaSet ): Integer;
var i: Integer;
begin
  i := Low( FAlphas );
  Result := -1;
  while ( Result < 0 ) and ( i <= High(FAlphas) ) do
  begin
    if a = FAlphas[ i ] then
      Result := i;
    Inc( i, 1 );
  end;
  if Result < 0 then
  begin
    SetLength( FAlphas, Length(FAlphas)+1 );
    Result := High(FAlphas);
    FAlphas[ Result ] := a;
  end;
end;

constructor TSPLexDFABuilder.Create( ldef: TLexerDefinition );
begin
  inherited;
  SetLength( FAlphas, 0 );
end;

procedure TSPLexDFABuilder.Generate;
var i, j: Integer;
    titer: TNFATransitionIterator;
begin
  if constructMinDFA then
  begin
    WriteLnO( 'unit '+LexerDefinition.LexerName+';' );
    WriteLnO( '' );
    WriteLnO( '{$IFDEF FPC}{$MODE DELPHI}{$ENDIF}{$H+}' );
    WriteLnO( '' );
    WriteLnO( 'interface' );
    WriteLnO( '' );
    WriteLnO( 'uses splex;' );
    WriteLnO( '' );
    WriteLnO( 'const ' );
    for i := 0 to CurrentNFA.FinalStateCount-1 do
      WriteLnO( '  '+CurrentNFA.FinalNames[ i ]+' = '+IntToStr( i+1 )+';' );
    WriteLnO( '' );
    WriteLnO( 'type' );
    WriteLnO( '  T'+LexerDefinition.LexerName+' = class( TSPLexScanDFA )' );
    WriteLnO( '    public constructor Create; override;' );
    WriteLnO( '  end;' );
    WriteLnO( '' );
    WriteLnO( 'implementation' );
    WriteLnO( '' );
    WriteLnO( 'constructor T'+LexerDefinition.LexerName+'.Create;' );
    WriteLnO( 'begin' );
    WriteLnO( '  FInitialNode := 0;' );
    WriteLnO( '  SetLength( FNodeTable, '+IntToStr( Length( CurrentNFA.Nodes ) )+' );' );
    for i := Low( CurrentNFA.Nodes ) to High( CurrentNFA.Nodes ) do
    begin
      WriteLnO( '  with FNodeTable[ '+IntToStr( i )+' ] do begin ' );
      if CurrentNFA.FinalState[ i ] >= TFAFinalMin then
        WriteLnO( '    state := '+CurrentNFA.FinalNames[ CurrentNFA.FinalState[ i ] ]+';' )
      else
        WriteLnO( '    state := L_NOSTATE;' );
      titer := CurrentNFA.Nodes[ i ].TransIterator;
      j := 0;
      while titer.Next do Inc( j, 1 );
      titer.Free;
      WriteLnO( '    SetLength( delta, '+IntToStr( j )+' );' );
      if j > 0 then
      begin
        titer := CurrentNFA.Nodes[ i ].TransIterator;
        j := 0;
        while titer.Next do
        begin
          WriteLnO( '    with delta[ '+IntToStr( j )+' ] do begin ' );
          WriteLnO( '      sigma := '+IntToStr( getAlphaRef( titer.CurrentTransSet ) )+';' );
          WriteLnO( '      dnode := '+IntToStr( titer.CurrentDestNode )+';' );
          WriteLnO( '    end;' );
          Inc( j, 1 );
        end;
        titer.Free;
      end;
      WriteLnO( '  end;' );
    end;
    WriteLnO( '  SetLength( FSigmaTable, '+IntToStr( Length( FAlphas ) )+' );' );
    for i := Low( FAlphas ) to High( FAlphas ) do
      WriteLnO( '  FSigmaTable[ '+IntToStr( i )+' ] := ['+ AlphaSetToString( FAlphas[ i ] ) +'];' );
    WriteLnO( 'end;' );
    WriteLnO( '' );
    WriteLnO( 'end.' );
  end;
end;


{ TSPLexDynamicDFABuilder }

constructor TSPLexDynamicDFABuilder.Create(ldef: TLexerDefinition);
begin
  inherited Create(ldef);
  FSPLexDFA := nil;
end;

procedure TSPLexDynamicDFABuilder.Generate;
var i, j: Integer;
    titer: TNFATransitionIterator;
begin
  FSPLexDFA := TSPLexDynamicScanDFA.Create;
  if constructMinDFA then
  begin
    // name mapping
    FSPLexDFA.SetFinals( CurrentNFA.FinalStateCount+1 );
    FSPLexDFA.SetFinalName( 0, '_EndOfStream' );  // add eos for name lookup
    for i := 1 to CurrentNFA.FinalStateCount do
      FSPLexDFA.SetFinalName( i, CurrentNFA.FinalNames[ i-1 ] );
    FSPLexDFA.SetNodeTableLength( Length( CurrentNFA.Nodes ) );
    for i := Low( CurrentNFA.Nodes ) to High( CurrentNFA.Nodes ) do
    begin
      if CurrentNFA.FinalState[ i ] >= TFAFinalMin then
        FSPLexDFA.SetNodeState( i, CurrentNFA.FinalState[ i ]+1 )
      else
        FSPLexDFA.SetNodeState( i, L_NOSTATE );
      titer := CurrentNFA.Nodes[ i ].TransIterator;
      j := 0;
      while titer.Next do Inc( j, 1 );
      titer.Free;
      FSPLexDFA.SetNodeDeltas( i, j );
      if j > 0 then
      begin
        titer := CurrentNFA.Nodes[ i ].TransIterator;
        j := 0;
        while titer.Next do
        begin
          FSPLexDFA.SetNodeDeltaSigma( i, j, getAlphaRef( titer.CurrentTransSet ) );
          FSPLexDFA.SetNodeDeltaDest( i, j, titer.CurrentDestNode );
          Inc( j, 1 );
        end;
        titer.Free;
      end;
    end;
    FSPLexDFA.SetSigmaTableLength( Length( FAlphas ) );
    for i := Low( FAlphas ) to High( FAlphas ) do
      FSPLexDFA.SetSigma( i, TSPLexSigmaSet( FAlphas[ i ] ) );
  end;
end;

destructor TSPLexDynamicDFABuilder.Destroy;
begin
  inherited Destroy;
  if FSPLexDFA <> nil then FSPLexDFA.Free;
end;

{ TSPLexDynamicScanDFA }

procedure TSPLexDynamicScanDFA.SetFinals(i: Integer);
begin
  SetLength( FFinals, i );
end;

procedure TSPLexDynamicScanDFA.SetFinalName(i: Integer; const s: String);
begin
  FFinals[ i ] := s;
end;

procedure TSPLexDynamicScanDFA.SetNodeTableLength(i: Integer);
begin
  SetLength( FNodeTable, i );
end;

procedure TSPLexDynamicScanDFA.SetNodeState(i, s: Integer);
begin
  FNodeTable[ i ].state := s;
end;

procedure TSPLexDynamicScanDFA.SetNodeDeltas(i, d: Integer);
begin
  SetLength( FNodeTable[ i ].delta, d );
end;

procedure TSPLexDynamicScanDFA.SetNodeDeltaSigma(i, d, aref: Integer);
begin
  FNodeTable[ i ].delta[ d ].sigma := aref;
end;

procedure TSPLexDynamicScanDFA.SetNodeDeltaDest(i, d, t: Integer);
begin
  FNodeTable[ i ].delta[ d ].dnode := t;
end;

procedure TSPLexDynamicScanDFA.SetSigmaTableLength(i: Integer);
begin
  SetLength( FSigmaTable, i );
end;

procedure TSPLexDynamicScanDFA.SetSigma(i: Integer; const s: TSPLexSigmaSet);
begin
  FSigmaTable[ i ] := s;
end;

function TSPLexDynamicScanDFA.GetFinalName(i: Integer): String;
begin
  Result := FFinals[ i ];
end;

constructor TSPLexDynamicScanDFA.Create;
begin
  FInitialNode := 0;
  inherited;
  SetLength( FFinals, 0 );
  SetLength( FNodeTable, 0 );
  SetLength( FSigmaTable, 0 );
end;

destructor TSPLexDynamicScanDFA.Destroy;
begin
  SetLength( FFinals, 0 );
  inherited Destroy;
end;

end.

