unit relant;

{$MODE DELPHI}{$H+}

//{$DEFINE DEBUG}

{$IFDEF DEBUG}
  {$ASSERTIONS ON}
{$ENDIF}

interface

uses sysutils;

type
  TTransformationException = class( Exception )
  end;

(******************************************************************************
 * Types and Constants used for RE->NFA->DFA->MinDFA Transformation
 ******************************************************************************)

  TFANodeType = Integer;
  TFAFinalType = Integer;
  TAlphaSet = set of Byte;
  TPartitionID = TFANodeType;

const
  TFANodeMin: TFANodeType = 0;
  TFANoNode: TFANodeType = -1;

  TFAFinalMin: TFAFinalType = 0;
  TFANoFinal: TFAFinalType = -1;
  TFAMultipleFinal: TFAFinalType = -2;

  TEmptyAlphaSet: TAlphaSet = [];

  TPartitionIDMin: TPartitionID = 0; // TFANodeMin;
  TPartitionPseudoID: TPartitionID = -1; // TFANoNode;

type
  TNFATransition = record
    DestNode: Integer;
    Transiton: TAlphaSet;
  end;

  PNFATransitionListElement = ^TNFATransitionListElement;
  TNFATransitionListElement = record
    Next: PNFATransitionListElement;
    Trans: TNFATransition;
  end;

  { TNFATransitionIterator }

  TNFATransitionIterator = class
    private
      FFront: PNFATransitionListElement;
      FCurrent: PNFATransitionListElement;
    public
      constructor Create( AFront: PNFATransitionListElement );
      function Next: Boolean;
      function CurrentTransSet: TAlphaSet;
      function CurrentDestNode: TFANodeType;
  end;

  { TNFATransitions }

  TNFATransitions = class
    private
      FHead: PNFATransitionListElement;
    public
      procedure AddTransition( ADestNode: TFANodeType;
                               const ATransSet: TAlphaSet );
      constructor Create;
      destructor Destroy; override;
      function Iterator: TNFATransitionIterator;
  end;

  { TNFANode }

  TNFANode = class
    private
      FTrans: TNFATransitions;
    public
      procedure AddTransitionN( ADestNode: TFANodeType;
                                const ATransSet: TAlphaSet );
      function AddTransitionD( ADestNode: TFANodeType;
                               const ATransSet: TAlphaSet ): Boolean;
      function TransIterator: TNFATransitionIterator;
      constructor Create;
      destructor Destroy; override;
  end;

  TNFANodes = array of TNFANode;

  { TNFADeltaS }

  TNFADeltaS = class
    private
      FNodes: TNFANodes;
      function getNodeCount: Integer;
    protected
      FStartNode: TFANodeType;
      FDeterministic: Boolean;
      FMin: Boolean;
      function addNode( n: TNFANode ): TFANodeType;
    public
      property Nodes: TNFANodes read FNodes;
      property NodeCount: Integer read getNodeCount;
      property StartNode: TFANodeType read FStartNode;
      property ConstructedDFA: Boolean read FDeterministic;
      property ConstructedMinDFA: Boolean read FMin;
      constructor Create; virtual;
      destructor Destroy; override;
  end;

  TFinalState = String;

  { TUNFA }

  TUNFA = class( TNFADeltaS )
    private
      FFinalName: TFinalState;
    protected
      FFinalNode: TFANodeType;
    public
      property FinalName: TFinalState read FFinalName;
      property FinalNode: TFANodeType read FFinalNode;
      constructor CreateEmpty( const AFinalName: TFinalState );
      constructor Create( const AFinalName: TFinalState;
                          const ASet: TAlphaSet ); reintroduce;
      destructor Destroy; override;
  end;

  TFinalStates = array of String;

  { TNFAFinalStates }

  TNFAFinalStates = class
    private
      FFinalStates: TFinalStates;
      function getFinalNum: Integer;
    public
      constructor Create;
      destructor Destroy; override;
      property Count: Integer read getFinalNum;
      property FinalStates: TFinalStates read FFinalStates;
      function AddFinalState( const AFinalState: TFinalState ): TFAFinalType;
  end;

  TNodeFinals = array of TFAFinalType;

  { TNFA }

  TNFA = class( TNFADeltaS )
    private
      FFinals: TNFAFinalStates;
      FNodeFinals: TNodeFinals;
      function getFinal( i: TFAFinalType ): TFinalState;
      function getFinalState(  n: TFANodeType ): TFAFinalType;
      function addNode( n: TNFANode; const f: TFinalState ): TFANodeType;
      function getFStateCount: Integer;
    public
      property FinalNames[ i: TFAFinalType ]: TFinalState read getFinal;
      property FinalState[ n: TFANodeType ]: TFAFinalType read getFinalState;
      property FinalStateCount: Integer read getFStateCount;
      procedure AddUNFA( const AUNFA: TUNFA );
      constructor Create; override;
      destructor Destroy; override;
  end;

  TREElementType = ( reoUNFA,     // UNFA -> Set/Char or Pattern representing NFA
                     reoMaxOne,   // ? - one or none times
                     reoMinOne,   // + - Kleene Plus -> multiple times, at least one
                     reoRepeat,   // * - Kleene Star -> multiple times
                     reoPatStart, // ( - (Sub)Pattern Start
                     reoPatEnd,   // ) - (Sub)Pattern End
                     reoOr        // | - or two Patterns
                   );

  TRegExElement = record
    Operation: TREElementType;
    LangSet: TUNFA;
  end;

  TRegExElements = array of TRegExElement;

  { TRegExString }

  TRegExString = class
     private
       FElements: TRegExElements;
       FFinalName: TFinalState;
       procedure addElement( AElement: TREElementType;
                             AUNFA: TUNFA );
     public
       property FinalName: TFinalState read FFinalName;
       constructor Create( const AFinalName: TFinalState );
       destructor Destroy; override;
       procedure AddUNFA( AUNFA: TUNFA );
       procedure AddOperation( AElement: TREElementType );
       function ToString: String;
   end;

(******************************************************************************
 * Regular Expression to UNFA Transformation
 ******************************************************************************)

  TUNFAUOps = ( uopMaxOne, uopKleeneStar, uopKleenePlus );

  TUNFAOperatorStackMarks = ( markFrame, markOpOr, markunfa );

  TUNFAOperatorElement = record
     mark: TUNFAOperatorStackMarks;
     unfa: TUNFA;
  end;

  TUNFAOperatorElements = array of TUNFAOperatorElement;

  { TUNFAOperator }

  TUNFAOperator = class
    private
      FEpsStack: TUNFAOperatorElements;
      FTop: Integer;
      procedure reInitialize;
      // push/pop
      procedure grow;
      procedure shrink;
      // stack operations
      procedure orTops;
      procedure catTops;
      // ( pushdown ) operations
      procedure pushFrameStart;
      procedure pushFrameEnd;
      procedure pushNFA( const AUNFA: TUNFA );
      procedure pushOr;
      procedure pushUOp( AUOp: TUNFAUOps );
      // internal Result
      // Top NFA -> after processing, this function returns ONLY ONCE
      // the Resulting NFA, as Long as Operation was completed without problems
      function resultNFA: TUNFA;
    public
      constructor Create;
      destructor Destroy; override;
      function RegExToUNFA( const AFinalName: String;
                            const ARegExString: TRegExString ): TUNFA;
      function RegExToOpString( const ARegExString: TRegExString ): String;
  end;

(******************************************************************************
 * NFA to DFA Transformation
 ******************************************************************************)

  PAlphaPartitionElement = ^TAlphaPartitionElement;
  TAlphaPartitionElement = record
    next: PAlphaPartitionElement;
    alphaset: TAlphaSet;
  end;

  { TAlphaSetPartitionIterator }

  TAlphaSetPartitionIterator = class
    private
      FFront: PAlphaPartitionElement;
      FCurrent: PAlphaPartitionElement;
    public
      constructor Create( AFront: PAlphaPartitionElement );
      function Next: Boolean;
      function CurrentSet: TAlphaSet;
  end;

  { TAlphaSetPartition }

  TAlphaSetPartition = class
    private
      FFront: PAlphaPartitionElement;
      procedure frontAddSet( {copy} AAlphaSet: TAlphaSet );
    public
      constructor Create;
      destructor Destroy; override;
      procedure AddAlphaSet( {copy} AAlphaSet: TAlphaSet );
      function Iterator: TAlphaSetPartitionIterator;
  end;

  TNodeSetFieldBlock = ValUInt;

  TNodeSetField = array of TNodeSetFieldBlock;

  { TNodeSetIterator }

  TNodeSetIterator = class
    private
      FBVectorRef: TNodeSetField;
      FSize: TFANodeType;
      FMask: TNodeSetFieldBlock;
      FAPos: Integer;
      FCurrent: TFANodeType;
    public
      constructor Create( const ABVectorRef: TNodeSetField;
                          ABSize: TFANodeType );
      procedure Reset;
      function Next: Boolean;
      function CurrentNode: TFANodeType;
  end;

  { TNodeSet }

  TNodeSet = class
    private
      FBVector: TNodeSetField;
      FBSize: TFANodeType;
      FHash: TNodeSetFieldBlock;
    public
      constructor Create( ABSize: TFANodeType );
      destructor Destroy; override;
      function SetNode( n: TFANodeType; AIsIn: Boolean ): Boolean;
      function IsEqual( ANodeSet: TNodeSet ): Boolean;
      function Iterator: TNodeSetIterator;
  end;

  TNodeSets = array of TNodeSet;

  { TNodeSetMap }

  TNodeSetMap = class
    private
      FNodeSets: TNodeSets;
      function mapNewNodeSet( const ANodeSet: TNodeSet ): TFANodeType;
    public
      property NodeSets: TNodeSets read FNodeSets;
      constructor Create;
      destructor Destroy; override;
      function MapUnifyNodeSet( var ANodeSet: TNodeSet ): TFANodeType;
  end;

  { TDFAConstructor }

  // what todo on Multiple Finalstates in One DFA Node --
  TDFACMultiStateResolution = ( msrInputOrder, // set first final in input order
                                               // for two languages L(R1), L(R2)
                                               // with L(R1) * L(R2) <> []
                                               // it creates
                                               // L(R2) := L(R2) - L(R1)
                                               // can result in empty Language
                                               // -> finalstate not in DFA
                                msrException );

  TDFAConstructor = class
    private
      FInputNFA: TNFA;
      FNodeMap: TNodeSetMap;
      FTransSet: TAlphaSetPartition;
      FAlpha: TAlphaSetPartitionIterator;
      FDFA: TNFA;
      FMultiStateResolve: TDFACMultiStateResolution;
      procedure reInitialize;
      function getFinalState( const ANFANodeSet: TNodeSet ): TFAFinalType;
      function getNoMulFinalState( const ANFANodeSet: TNodeSet ): TFAFinalType;
      function getOrderFinalState( const ANFANodeSet: TNodeSet ): TFAFinalType;
      procedure addDFANode( ANode: TFANodeType; AFinalState: TFAFinalType );
      procedure resolveEpsilons( var ANFANodeSet: TNodeSet );
      function getTransitionPartition(
        const ANFANodeSet: TNodeSet ): TAlphaSetPartition;
      function resolveNode( const AStart: TNodeSet;
        const ATransitionSet: TAlphaSet ): TFANodeType;
      procedure buildDFAInternals;
    public
      property MultiStateResolution: TDFACMultiStateResolution
        read FMultiStateResolve
        write FMultiStateResolve;
      constructor Create;
      destructor Destroy; override;
      function ConstructDFA( const AInputNFA: TNFA ): TNFA;
  end;

  (******************************************************************************
   * DFA to MinDFA Transformation
   ******************************************************************************)

  TNodeRefField = array of TFANodeType;

  { TNodePartition }

  TNodePartition = class
    private
      FPartID: TPartitionID;
      FNodeRefs: TNodeRefField;
    public
      property PartitionID: TPartitionID read FPartID;
      property NodeRefs: TNodeRefField read FNodeRefs;
      constructor Create( APartID: TPartitionID );
      destructor Destroy; override;
      procedure AddNode( ANode: TFANodeType );
  end;

  TNodePartitions = array of TNodePartition;

  { TNodePartitionSet }

  TNodePartitionSet = class
    private
      FParts: TNodePartitions;
      procedure addPartition( APart: TNodePartition );
    public
      property Partitions: TNodePartitions read FParts;
      constructor Create;
      destructor Destroy; override;
      function GetPartition( APartID: TPartitionID ): TNodePartition;
      procedure AddNode( APartID: TPartitionID; ANodeID: TFANodeType );
  end;

  TPartitionIDs = array of TPartitionID;

  { TNodePartitionMap }

  TNodePartitionMap = class
    private
      FPartIDs: TPartitionIDs;
      FParts: TNodePartitionSet;
      FTopPart: TPartitionID;
      function getPartCount: Integer;
      procedure remapNodesForPart( APart: TNodePartition );
    public
      property PartitionCount: Integer read getPartCount;
      constructor Create( ANodeNum: Integer );
      destructor Destroy; override;
      function PartIDLookup( ANodeId: TFANodeType ): TPartitionID;
      function GetPartition( APartID: TPartitionID ): TNodePartition;
      function RefinePartition( ARefPartID: TPartitionID;
                                var ASubPartition: TNodePartitionSet ): Boolean;
  end;


  { TMinDFAConstructor }

  TMinDFAConstructor = class
  private
    FInputNFA: TNFA;
    FPartMap: TNodePartitionMap;
    FMinDFA: TNFA;
    procedure reInitialize;
    function resolveTransSet( APart: TNodePartition ): TAlphaSetPartition;
    function resolveDestPartSet( const AAlpha: TAlphaSet;
                                 APart: TNodePartition ): TNodePartitionSet;
    procedure buildNodePartition;
    procedure buildMinDFA;
  public
    constructor Create;
    destructor Destroy; override;
    function ConstructMinDFA( const AInputNFA: TNFA ): TNFA;
  end;

// RE To NFA Transformation
function RegExToUNFA( const AFinalName: String;
                      const ARegExString: TRegExString ): TUNFA;
function RegExToOpString( const ARegExString: TRegExString ): String;

// NFA To DFA Transformation
procedure AddUNFAToNFA( const AUENFA: TUNFA; ANFA: TNFA );
function BuildDFA(const ANFA: TNFA ): TNFA;

// DFA To DFA Transformation (minimization)
function BuildMinDFA(const ADFA: TNFA ): TNFA;

// verifications
  // verify NodeTable/FinalStateMap and Transition consistency
  // check wether this is an NFA or a DFA
function VerifyDFA(const ADFA: TNFA ): Boolean;
  // verify NodeTable/FinalStateMap and Transition consistency
  //  -> No Checks for redundancy!
function VerifyNFA(const ANFA: TNFA ): Boolean;

// Stats
function DeltaNodeCount( const d: TNFADeltaS ): Integer;
function DeltaTransCount( const d: TNFADeltaS ): Integer;


implementation

function RegExToUNFA(const AFinalName: String;
                     const ARegExString: TRegExString): TUNFA;
var unfaops: TUNFAOperator;
begin
  unfaops := TUNFAOperator.Create;
  try
    Result := unfaops.RegExToUNFA( AFinalName, ARegExString );
  finally
    unfaops.Free;
  end;
end;

function RegExToOpString(const ARegExString: TRegExString): String;
var unfaops: TUNFAOperator;
begin
  unfaops := TUNFAOperator.Create;
  try
    Result := unfaops.RegExToOpString( ARegExString );
  finally
    unfaops.Free;
  end;
end;

procedure AddUNFAToNFA(const AUENFA: TUNFA; ANFA: TNFA);
begin
  ANFA.AddUNFA( AUENFA );
end;

function BuildDFA(const ANFA: TNFA): TNFA;
var dfabuilder: TDFAConstructor;
begin
  dfabuilder := TDFAConstructor.Create;
  try
    Result := dfabuilder.ConstructDFA( ANFA );
  finally
    dfabuilder.Free;
  end;
end;

function BuildMinDFA(const ADFA: TNFA): TNFA;
var dfabuilder: TMinDFAConstructor;
begin
  dfabuilder := TMinDFAConstructor.Create;
  try
    Result := dfabuilder.ConstructMinDFA( ADFA );
  finally
    dfabuilder.Free;
  end;
end;

function VerifyDFA(const ADFA: TNFA): Boolean;
var i: Integer;
    titer: TNFATransitionIterator;
    talphas: TAlphaSet;
begin
  // consistency in structure?
  Result := VerifyNFA( ADFA );
  // flagged deterministic?
  Result := Result and ADFA.FDeterministic;
  if Result then
  begin
    // check all transitions for one node are disjoint subsets for this node
    i := Low( ADFA.FNodes );
    while Result and ( i <= High( ADFA.FNodes ) ) do
    begin
      titer := ADFA.FNodes[ i ].TransIterator;
      talphas := TEmptyAlphaSet;
      while Result and titer.Next do
      begin
        // all sets must be non empty, and all sets are partition subsets (disjoint)
        Result := ( titer.CurrentTransSet <> TEmptyAlphaSet ) and
                  ( ( titer.CurrentTransSet * talphas ) = TEmptyAlphaSet );
        talphas := talphas + titer.CurrentTransSet;
      end;
      titer.Free;
      Inc( i, 1 );
    end;
  end;
end;

function VerifyNFA(const ANFA: TNFA): Boolean;
var i: Integer;
    titer: TNFATransitionIterator;
begin
  // check wether nodes exists and startnode is set, or
  // no nodes and startnode is TFANoNode
  Result := ( ( ANFA.FStartNode >= TFANodeMin ) and
              ( ANFA.FStartNode < ANFA.getNodeCount ) ) or
            ( ( ANFA.FStartNode = TFANoNode ) and
              ( ANFA.getNodeCount = 0 ) );
  // check FNodes size against FinalState map for Nodes
  Result := Result and ( ANFA.getNodeCount = Length( ANFA.FNodeFinals ) );
  if Result then
  begin
    // check all finalstate ids are in the range of the finalstates map or
    //       flagged NonFinal (TFANoFinal)
    // check all DestNodes for Transitions are in range of FNodes
    i := Low( ANFA.FNodes );
    while Result and ( i <= High( ANFA.FNodes ) ) do
    begin
      Result := ( ANFA.FNodeFinals[ i ] = TFANoFinal ) or
                ( ANFA.FNodeFinals[ i ] < Length( ANFA.FFinals.FFinalStates ) );
      if Result then
      begin
        titer := ANFA.FNodes[ i ].TransIterator;
        while Result and titer.Next do
        begin
          Result := ( titer.CurrentDestNode >= TFANodeMin ) and
                    ( titer.CurrentDestNode < ANFA.getNodeCount );
        end;
        titer.Free;
      end;
      Inc( i, 1 );
    end;
  end;
end;

function DeltaNodeCount( const d: TNFADeltaS ): Integer;
begin
  Result := d.getNodeCount;
end;

function DeltaTransCount( const d: TNFADeltaS ): Integer;
var i: Integer;
    titer : TNFATransitionIterator;
begin
  Result := 0;
  for i := 0 to d.getNodeCount-1 do
  begin
    titer := d.Nodes[ i ].FTrans.Iterator;
    while titer.Next do Inc( Result, 1 );
    titer.Free;
  end;
end;

{ TNFATransitionIterator }

constructor TNFATransitionIterator.Create(AFront: PNFATransitionListElement);
begin
  FFront := AFront;
  FCurrent := nil;
end;

function TNFATransitionIterator.Next: Boolean;
begin
  if FCurrent <> nil then
  begin
    Result := FCurrent^.Next <> nil;
    if Result then FCurrent := FCurrent^.Next;
  end else
  begin
    FCurrent := FFront;
    Result := FCurrent <> nil;
  end;
end;

function TNFATransitionIterator.CurrentTransSet: TAlphaSet;
begin
  Result := FCurrent^.Trans.Transiton;
end;

function TNFATransitionIterator.CurrentDestNode: TFANodeType;
begin
  Result := FCurrent^.Trans.DestNode;
end;

{ TNFATransitions }

procedure TNFATransitions.AddTransition(ADestNode: TFANodeType;
  const ATransSet: TAlphaSet);
var tmptrans: PNFATransitionListElement;
begin
  tmptrans := FHead;
  while tmptrans <> nil do
  begin
    if tmptrans^.Trans.DestNode = ADestNode then
    begin
      // merge transitions with same target
      tmptrans^.Trans.Transiton :=
        tmptrans^.Trans.Transiton + ATransSet;
      break;
    end;
    tmptrans := tmptrans^.Next;
  end;
  if tmptrans = nil then
  begin
    tmptrans := New( PNFATransitionListElement );
    tmptrans^.Next := FHead;
    FHead := tmptrans;
    tmptrans^.Trans.DestNode := ADestNode;
    tmptrans^.Trans.Transiton := ATransSet;
  end;
end;

constructor TNFATransitions.Create;
begin
  FHead := nil;
end;

destructor TNFATransitions.Destroy;
var tmptrans: PNFATransitionListElement;
begin
  while FHead <> nil do
  begin
    tmptrans := FHead^.Next;
    Dispose( FHead );
    FHead := tmptrans;
  end;
  inherited Destroy;
end;

function TNFATransitions.Iterator: TNFATransitionIterator;
begin
  Result := TNFATransitionIterator.Create( FHead );
end;

{ TNFANode }

procedure TNFANode.AddTransitionN(ADestNode: TFANodeType;
  const ATransSet: TAlphaSet);
begin
  FTrans.AddTransition( ADestNode, ATransSet );
end;

function TNFANode.AddTransitionD(ADestNode: TFANodeType;
  const ATransSet: TAlphaSet): Boolean;
var iter: TNFATransitionIterator;
begin
  iter := TransIterator;
  Result := true;
  while Result and iter.Next do
  begin
    Result := ( iter.CurrentDestNode = ADestNode ) or
              ( ( iter.CurrentTransSet * ATransSet ) = TEmptyAlphaSet );
  end;
  iter.Free;
  if Result then
    FTrans.AddTransition( ADestNode, ATransSet );
end;

function TNFANode.TransIterator: TNFATransitionIterator;
begin
  Result := FTrans.Iterator;
end;

constructor TNFANode.Create;
begin
  FTrans := TNFATransitions.Create;
end;

destructor TNFANode.Destroy;
begin
  FTrans.Free;
  inherited Destroy;
end;

{ TNFADeltaS }

function TNFADeltaS.getNodeCount: Integer;
begin
  Result := Length( FNodes );
end;

function TNFADeltaS.addNode(n: TNFANode): TFANodeType;
begin
  SetLength( FNodes, Succ( Length( FNodes ) ) );
  FNodes[ High( FNodes ) ] := n;
  Result := High( FNodes );
end;

constructor TNFADeltaS.Create;
begin
  SetLength( FNodes, 0 );
  FDeterministic := false;
  FMin := false;
  FStartNode := TFANoNode;
end;

destructor TNFADeltaS.Destroy;
var i: Integer;
begin
  for i := Low( FNodes ) to High( FNodes ) do
    FNodes[ i ].Free;
  SetLength( FNodes, 0 );
  inherited Destroy;
end;

{ TUNFA }

constructor TUNFA.CreateEmpty(const AFinalName: TFinalState);
begin
  inherited Create;
  FFinalName := AFinalName;
  FFinalNode := TFANoNode;
end;

constructor TUNFA.Create(const AFinalName: TFinalState; const ASet: TAlphaSet);
begin
  inherited Create;
  FFinalName := AFinalName;
  FStartNode := addNode( TNFANode.Create );
  FFinalNode := addNode( TNFANode.Create );
  FNodes[ FStartNode ].AddTransitionN( FFinalNode, ASet );
end;

destructor TUNFA.Destroy;
begin
  inherited Destroy;
end;

{ TNFAFinalStates }

function TNFAFinalStates.getFinalNum: Integer;
begin
  Result := Length( FFinalStates );
end;

constructor TNFAFinalStates.Create;
begin
  SetLength( FFinalStates, 0 );
end;

destructor TNFAFinalStates.Destroy;
begin
  SetLength( FFinalStates, 0 );
  inherited Destroy;
end;

function TNFAFinalStates.AddFinalState(
  const AFinalState: TFinalState): TFAFinalType;
var i: Integer;
begin
  Result := TFANoFinal;
  if AFinalState <> '' then
  begin
    i := Low( FFinalStates );
    while ( Result = TFANoFinal ) and ( i <= High( FFinalStates ) ) do
    begin
      if FFinalStates[ i ] = AFinalState then
        Result := i;
      Inc( i, 1 );
    end;
    if Result = TFANoFinal then
    begin
      SetLength( FFinalStates, Succ( Length( FFinalStates ) ) );
      FFinalStates[ High( FFinalStates ) ] := AFinalState;
      Result := High( FFinalStates );
    end;
  end;
end;

{ TNFA }

function TNFA.getFinal(i: TFAFinalType): TFinalState;
begin
  Result := FFinals.FFinalStates[ i ];
end;

function TNFA.getFinalState(n: TFANodeType): TFAFinalType;
begin
  Result := FNodeFinals[ n ];
end;

function TNFA.addNode(n: TNFANode; const f: TFinalState): TFANodeType;
var fid: TFAFinalType;
begin
  fid := FFinals.AddFinalState( f );
  SetLength( FNodeFinals, Succ( Length( FNodeFinals ) ) );
  FNodeFinals[ High( FNodeFinals ) ] := fid;
  Result := inherited addNode( n );
end;

function TNFA.getFStateCount: Integer;
begin
  Result := FFinals.Count;
end;

procedure TNFA.AddUNFA(const AUNFA: TUNFA);
var ofs: TFANodeType;
    i: Integer;
    titer: TNFATransitionIterator;
begin
  if FStartNode <> TFANoNode then
  begin
    // yes.. this is highly unoptimized due to the possibility
    // of adding UNFAs to DFAs/NFAs which have a backlink to startnode or
    // non-epsilon transitions from the startnode -> this can build another
    // NFA (f.e. (abc) and (cde)* to ((abc)|(cde))* instead of (abc)|((cde)*) )
    // either -> allow only unfas and make one startnode with multiple e-trans
    // or -> introduce backlinking in nfanodes, check startnode transitions
    //       for non-epsilon transitions and backlinks, if either of them exists
    //       create a new node (has some other advantage too)
    ofs := addNode( TNFANode.Create, '' );
    FNodes[ ofs ].AddTransitionN( FStartNode, TEmptyAlphaSet );
    FStartNode := ofs;
  end else
  begin
    FStartNode := addNode( TNFANode.Create, '' );
  end;
  // copy nodes
  ofs := getNodeCount;
  SetLength( FNodes, Length( FNodes ) + AUNFA.getNodeCount );
  for i := Low( AUNFA.FNodes ) to High( AUNFA.FNodes ) do
  begin
    FNodes[ i+ofs ] := TNFANode.Create;
    titer := AUNFA.FNodes[ i ].TransIterator;
    while titer.Next do
    begin
      FNodes[ i+ofs ].
        AddTransitionN( titer.CurrentDestNode + ofs, titer.CurrentTransSet );
    end;
    titer.Free;
  end;
  // links startnode
  FNodes[ FStartNode ].AddTransitionN( AUNFA.StartNode+ofs, TEmptyAlphaSet );
  // set Finals
  SetLength( FNodeFinals, Length( FNodeFinals ) + AUNFA.getNodeCount );
  for i := ofs to High( FNodeFinals ) do
    FNodeFinals[ i ] := TFANoFinal;
  FNodeFinals[ AUNFA.FinalNode + ofs ] :=
    FFinals.AddFinalState( AUNFA.FinalName );
end;

constructor TNFA.Create;
begin
  inherited Create;
  FFinals := TNFAFinalStates.Create;
  SetLength( FNodeFinals, 0 );
end;

destructor TNFA.Destroy;
begin
  FFinals.Free;
  SetLength( FNodeFinals, 0 );
  inherited Destroy;
end;

{ TRegExString }

procedure TRegExString.addElement(AElement: TREElementType; AUNFA: TUNFA);
begin
  SetLength( FElements, Succ( Length( FElements ) ) );
  FElements[ High( FElements ) ].Operation := AElement;
  FElements[ High( FElements ) ].LangSet := AUNFA;
end;

constructor TRegExString.Create(const AFinalName: TFinalState);
begin
  FFinalName := AFinalName;
  SetLength( FElements, 0 );
end;

destructor TRegExString.Destroy;
var i: Integer;
begin
  for i := Low( FElements ) to High( FElements ) do
    if FElements[ i ].LangSet <> nil then
      FElements[ i ].LangSet.Free;
  inherited Destroy;
end;

procedure TRegExString.AddUNFA(AUNFA: TUNFA);
begin
  if AUNFA <> nil then
    addElement( reoUNFA, AUNFA )
  else
    addElement( reoUNFA, TUNFA.Create( '[]', TEmptyAlphaSet ) );
end;

procedure TRegExString.AddOperation(AElement: TREElementType);
begin
  if AElement <> reoUNFA then
  begin
    addElement( AElement, nil );
  end else AddUNFA( nil );
end;

function TRegExString.ToString: String;
var i: Integer;
begin
  Result := '';
  for i := Low( FElements ) to High( FElements ) do
  begin
    case FElements[ i ].Operation of
      reoMaxOne: Result := Result + '?';
      reoMinOne: Result := Result + '+';
      reoPatEnd: Result := Result + ')';
      reoPatStart: Result := Result + '(';
      reoRepeat: Result := Result + '*';
      reoOr: Result := Result + '|';
      reoUNFA: Result := Result + FElements[ i ].LangSet.FFinalName;
      else
        Result := Result + '<ERROR>';
    end;
  end;
end;

(******************************************************************************
 * Regular Expression to UNFA Transformation
 ******************************************************************************)

procedure ofscopynodes( ofs: Integer; const src: TUNFA; var dest: TUNFA );
var i: Integer;
    cnode: TNFANode;
    iter: TNFATransitionIterator;
begin
  for i := 0 to src.NodeCount-1 do
  begin
    cnode := TNFANode.Create;
    iter := src.Nodes[ i ].TransIterator;
    while iter.Next do
      cnode.AddTransitionN( iter.CurrentDestNode+ofs, iter.CurrentTransSet );
    iter.Free;
    dest.addNode( cnode );
  end;
end;

function mergecopyreindex( const LOp, ROp: TUNFA; var dest: TUNFA ): Integer;
begin
  Assert( dest.NodeCount = 0 );
  Result := LOp.NodeCount;
  ofscopynodes( 0, LOp, dest );
  ofscopynodes( Result, ROp, dest );
end;

// unfa/re transformation operations
function UNFAOpOr( const LOp, ROp: TUNFA ): TUNFA;
var ofs: Integer;
begin
  Assert( LOp.NodeCount <> 0, 'Left Operand is Empty' );
  Assert( ROp.NodeCount <> 0, 'Right Operand is Empty' );
  Result := TUNFA.CreateEmpty( '(' + LOp.FFinalName + '|' +
                                     ROp.FFinalName + ')' );
  // merge LOP nodes from 0..ofs-1 and ROp nodes from ofs..newcount
  // reindex the transitions for the new node places
  ofs := mergecopyreindex( LOp, ROp, Result );
  // add transitions from new start to old starts
  Result.FStartNode := Result.addNode( TNFANode.Create );
  Result.Nodes[ Result.FStartNode ].
    AddTransitionN( LOp.StartNode, TEmptyAlphaSet );
  Result.Nodes[ Result.FStartNode ].
    AddTransitionN( ROp.StartNode+ofs, TEmptyAlphaSet );
  // add transitions from old finals to new final
  Result.FFinalNode := Result.addNode( TNFANode.Create );
  Result.Nodes[ LOp.FinalNode ].
    AddTransitionN( Result.FinalNode, TEmptyAlphaSet );
  Result.Nodes[ ROp.FinalNode+ofs ].
    AddTransitionN( Result.FinalNode, TEmptyAlphaSet );
end;

function UNFAOpCon( const LOp, ROp: TUNFA ): TUNFA;
var ofs: Integer;
begin
  Assert( LOp.NodeCount <> 0, 'Left Operand is Empty' );
  Assert( ROp.NodeCount <> 0, 'Right Operand is Empty' );
  Result := TUNFA.CreateEmpty( '(' + LOp.FFinalName + '.' +
                                     ROp.FFinalName + ')' );
  // merge LOP nodes from 0..ofs-1 and ROp nodes from ofs..newcount
  // reindex the transitions for the new node places
  ofs := mergecopyreindex( LOp, ROp, Result );
  // add transitions from old final to old start
  Result.Nodes[ LOp.FinalNode ].
    AddTransitionN( ROp.StartNode+ofs, TEmptyAlphaSet );
  // set new final and start
  Result.FStartNode := LOp.StartNode;
  Result.FFinalNode := ROp.FinalNode+ofs;
end;

function UNFAOpMaxOne( const Op: TUNFA ): TUNFA;
begin
  Assert( Op.NodeCount <> 0, 'Operand is Empty' );
  Result := TUNFA.CreateEmpty( '(' + Op.FinalName + '?)' );
  ofscopynodes( 0, Op, Result );
  // create new start and end
  Result.FStartNode := Result.addNode( TNFANode.Create );
  Result.FFinalNode := Result.addNode( TNFANode.Create );
  // connect starts and ends, such that Result is the "same" as Op
  Result.Nodes[ Result.StartNode ].
    AddTransitionN( Op.StartNode, TEmptyAlphaSet );
  Result.Nodes[ Op.FinalNode ].
    AddTransitionN( Result.FinalNode, TEmptyAlphaSet );
  // connect start with and final, which allows to hit this "None times"
  Result.Nodes[ Result.StartNode ].
    AddTransitionN( Result.FinalNode, TEmptyAlphaSet );
end;

function UNFAOpKleenePlus( const Op: TUNFA ): TUNFA;
begin
  Assert( Op.NodeCount <> 0, 'Operand is Empty' );
  Result := TUNFA.CreateEmpty( '(' + Op.FinalName + '+)' );
  ofscopynodes( 0, Op, Result );
  // create new start and end
  Result.FStartNode := Result.addNode( TNFANode.Create );
  Result.FFinalNode := Result.addNode( TNFANode.Create );
  // connect starts and ends, such that Result is the "same" as Op
  Result.Nodes[ Result.StartNode ].
    AddTransitionN( Op.StartNode, TEmptyAlphaSet );
  Result.Nodes[ Op.FinalNode ].
    AddTransitionN( Result.FinalNode, TEmptyAlphaSet );
  // connect final with and start, which allows to hit this multiple times
  Result.Nodes[ Result.FinalNode ].
    AddTransitionN( Result.StartNode, TEmptyAlphaSet );
end;

function UNFAOpKleeneStar( const Op: TUNFA ): TUNFA;
var tmpnfa: TUNFA;
begin
  Assert( Op.NodeCount <> 0, 'Operand is Empty' );
  // (R+)? -> (R|RR...)? -> []|(R|RR...) ~ R*
  tmpnfa := UNFAOpKleenePlus( OP );
  Result := UNFAOpMaxOne( tmpnfa );
  Result.FFinalName := '(' + Op.FinalName + '*)';
  tmpnfa.Free;
end;

{ TUNFAOperator }

procedure TUNFAOperator.reInitialize;
begin
  while FTop > 0 do
  begin
    if FEpsStack[ Pred( FTop ) ].unfa <> nil then
      FEpsStack[ Pred( FTop ) ].unfa.Free;
  end;
  SetLength( FEpsStack, 0 );
end;

procedure TUNFAOperator.grow;
begin
  Inc( FTop, 1 );
  if FTop < Length( FEpsStack ) then
    SetLength( FEpsStack, Length( FEpsStack )+16 );
  FEpsStack[ Pred( FTop ) ].unfa := nil;
end;

procedure TUNFAOperator.shrink;
begin
  if FTop > 0 then Dec( FTop, 1 );
end;

function TUNFAOperator.resultNFA: TUNFA;
begin
  if FTop <> 1 then
    raise TTransformationException.
            Create( 'More or Less then 1 Stack Element' );
  Result := FEpsStack[ Pred( FTop ) ].unfa;
  if Result <> nil then shrink;
end;

constructor TUNFAOperator.Create;
begin
  FTop := 0;
  SetLength( FEpsStack, 16 );
end;

destructor TUNFAOperator.Destroy;
begin
  reInitialize;
  inherited Destroy;
end;

function TUNFAOperator.RegExToUNFA( const AFinalName: String;
  const ARegExString: TRegExString): TUNFA;
var i: Integer;
begin
  try
    pushFrameStart;
    for i := Low( ARegExString.FElements ) to High( ARegExString.FElements ) do
    begin
      case ARegExString.FElements[ i ].Operation of
        reoMaxOne: pushUOp( uopMaxOne );
        reoMinOne: pushUOp( uopKleenePlus );
        reoPatEnd: pushFrameEnd;
        reoPatStart: pushFrameStart;
        reoRepeat: pushUOp( uopKleeneStar );
        reoOr: pushOr;
        reoUNFA: pushNFA( ARegExString.FElements[ i ].LangSet );
      end;
    end;
    pushFrameEnd;
    Result := resultNFA;
    Result.FFinalName := AFinalName;
  finally
    reInitialize;
  end;
end;

function TUNFAOperator.RegExToOpString(
  const ARegExString: TRegExString): String;
var i: Integer;
    tmpnfa: TUNFA;
begin
  try
    pushFrameStart;
    for i := Low( ARegExString.FElements ) to High( ARegExString.FElements ) do
    begin
      case ARegExString.FElements[ i ].Operation of
        reoMaxOne: pushUOp( uopMaxOne );
        reoMinOne: pushUOp( uopKleenePlus );
        reoPatEnd: pushFrameEnd;
        reoPatStart: pushFrameStart;
        reoRepeat: pushUOp( uopKleeneStar );
        reoOr: pushOr;
        reoUNFA: pushNFA( ARegExString.FElements[ i ].LangSet );
      end;
    end;
    pushFrameEnd;
    tmpnfa := resultNFA;
    Result := tmpnfa.FFinalName;
    tmpnfa.Free;
  finally
    reInitialize;
  end;
end;

procedure TUNFAOperator.pushFrameStart;
begin
  grow;
  FEpsStack[ Pred( FTop ) ].mark := markFrame;
end;

procedure TUNFAOperator.catTops;
var o,o2: TUNFA;
begin
  // @pre: FTop > 1 and top = unfa
  if FTop <= 1 then
    raise TTransformationException.
            Create( 'Stack Underflow' );
  if FEpsStack[ Pred( FTop ) ].mark <> markunfa then
    raise TTransformationException.
            Create( 'Wrong Left Operand for Concatenation' );
  o := nil;
  repeat
    if o <> nil then
    begin
      o2 := FEpsStack[ Pred( FTop ) ].unfa;
      FEpsStack[ Pred( FTop ) ].unfa := UNFAOpCon( o2, o );
      o2.Free;
      o.Free;
    end;
    o := FEpsStack[ Pred( FTop ) ].unfa;
    shrink;
  until ( FTop <= 0 ) or ( FEpsStack[ Pred( FTop ) ].mark <> markunfa );
  if FTop <= 0 then
  begin
    if o <> nil then o.Free;
    raise TTransformationException.
            Create( 'Stack Underflow' );
  end;
  grow;
  FEpsStack[ Pred( FTop ) ].unfa := o;
  FEpsStack[ Pred( FTop ) ].mark := markunfa;
  // @post: FTop > 1, top = unfa
end;

procedure TUNFAOperator.orTops;
var o, o2: TUNFA;
begin
  // @pre: ( ( FTop > 1 ) and ( top = OpOr ) )
  //    or ( ( FTop > 2 ) and ( top = unfa ) and ( pred(top) = OpOr ) )
  if FTop <= 1 then
    raise TTransformationException.
            Create( 'Stack Underflow' );
  // get right operand
  if FEpsStack[ Pred( FTop ) ].mark = markunfa then
  begin
    //   pred(top)=or and top=unfa -> pop(o1),pop(-)
    o := FEpsStack[ Pred( FTop ) ].unfa;
    shrink;
  end else
  begin
    // top = or -> pop(-), o1 = empty
    o := TUNFA.Create( '[]', TEmptyAlphaSet );
  end;
  // pop OpOr
  shrink;
  if FTop <= 1 then
  begin
    o.Free;
    raise TTransformationException.
            Create( 'Stack Underflow' );
  end;
  // get left Operand
  if FEpsStack[ Pred( FTop ) ].mark = markunfa then
  begin
    //   top = unfa -> catTops, pop->o2, or(o1,o2), push(o1)
    catTops;
    o2 := FEpsStack[ Pred( FTop ) ].unfa;
    FEpsStack[ Pred( FTop ) ].unfa := UNFAOpOr( o2, o );
    o2.Free;
    o.Free;
  end else
  begin
    //   top <> unfa -> ..(|R or (..||R -> R? -> o1.MaxOne, push(o1)
    o2 := o;
    o := UNFAOpMaxOne( o2 );
    o2.Free;
    grow;
    FEpsStack[ Pred( FTop ) ].mark := markunfa;
    FEpsStack[ Pred( FTop ) ].unfa := o;
  end;
  // @post: ( FTop >= 2 ) and ( top = unfa ) and ( pred(top) <> unfa )
end;

procedure TUNFAOperator.pushFrameEnd;
var o: TUNFA;
    done: Boolean;
begin
  done := false;
  while not done do
  begin
    done := FTop <= 0;
    if not done then
    begin
      case FEpsStack[ Pred( FTop ) ].mark of
        markFrame:
          begin
            // top = mark -> pop(-), push(empty) - done
            shrink;
            grow;
            FEpsStack[ Pred( FTop ) ].unfa :=
              TUNFA.Create( '[]', TEmptyAlphaSet );
            FEpsStack[ Pred( FTop ) ].mark := markunfa;
            done := true;
          end;
        markOpOr:
          // top = or
          orTops;
        markunfa:
          // top = unfa ->
          begin
            if FTop <= 1 then
              raise TTransformationException.
                      Create( 'Stack Underflow' );
            case FEpsStack[ Pred( Pred( FTop ) ) ].mark of
              markunfa:
                // pred(top)=unfa and top=unfa -> catTops
                catTops;
              markFrame:
                begin
                  // pred(top)=mark and top=unfa -> pop->o1,pop(-),push(o1)
                  // - done
                  o := FEpsStack[ Pred( FTop ) ].unfa;
                  shrink;
                  shrink;
                  grow;
                  FEpsStack[ Pred( FTop ) ].unfa := o;
                  FEpsStack[ Pred( FTop ) ].mark := markunfa;
                  done := true;
                end;
              markOpOr:
                // pred(top) = or and top = unfa
                orTops;
            end;
          end;
      end;
    end;
  end;
end;

procedure TUNFAOperator.pushNFA( const AUNFA: TUNFA );
begin
  grow;
  FEpsStack[ Pred( FTop ) ].unfa := TUNFA.CreateEmpty( AUNFA.FinalName );
  ofscopynodes( 0, AUNFA, FEpsStack[ Pred( FTop ) ].unfa );
  FEpsStack[ Pred( FTop ) ].unfa.FStartNode := AUNFA.FStartNode;
  FEpsStack[ Pred( FTop ) ].unfa.FFinalNode := AUNFA.FFinalNode;
  FEpsStack[ Pred( FTop ) ].mark := markunfa;
end;

procedure TUNFAOperator.pushOr;
begin
  grow;
  FEpsStack[ Pred( FTop ) ].mark := markOpOr;
end;

procedure TUNFAOperator.pushUOp( AUOp: TUNFAUOps );
var o: TUNFA;
begin
  if FEpsStack[ Pred( FTop ) ].unfa <> nil then
  begin
    o := FEpsStack[ Pred( FTop ) ].unfa;
    case AUOp of
      uopMaxOne: FEpsStack[ Pred( FTop ) ].unfa := UNFAOpMaxOne( o );
      uopKleeneStar: FEpsStack[ Pred( FTop ) ].unfa := UNFAOpKleeneStar( o );
      uopKleenePlus: FEpsStack[ Pred( FTop ) ].unfa := UNFAOpKleenePlus( o );
    end;
    if o <> FEpsStack[ Pred( FTop ) ].unfa then o.Free;
  end; // -> else, ignored operation on "empty word" NFA
end;

{ TAlphaSetPartitionIterator }

constructor TAlphaSetPartitionIterator.Create(AFront: PAlphaPartitionElement);
begin
  FFront := AFront;
  FCurrent := nil;
end;

function TAlphaSetPartitionIterator.Next: Boolean;
begin
  if FCurrent <> nil then
  begin
    Result := FCurrent^.Next <> nil;
    if Result then FCurrent := FCurrent^.Next;
  end else
  begin
    FCurrent := FFront;
    Result := FCurrent <> nil;
  end;
end;

function TAlphaSetPartitionIterator.CurrentSet: TAlphaSet;
begin
  Result := FCurrent^.alphaset;
end;

{ TAlphaSetPartition }

constructor TAlphaSetPartition.Create;
begin
  FFront := nil;
end;

destructor TAlphaSetPartition.Destroy;
var tmppart: PAlphaPartitionElement;
begin
  while FFront <> nil do
  begin
    tmppart := FFront^.Next;
    Dispose( FFront );
    FFront := tmppart;
  end;
end;

procedure TAlphaSetPartition.frontAddSet( AAlphaSet: TAlphaSet );
var tmppart: PAlphaPartitionElement;
begin
  tmppart := New( PAlphaPartitionElement );
  tmppart^.next := FFront;
  tmppart^.alphaset := AAlphaSet;
  FFront := tmppart;
end;

procedure TAlphaSetPartition.AddAlphaSet( AAlphaSet: TAlphaSet );
var i: PAlphaPartitionElement;
    t: TAlphaSet;
{$IFDEF DEBUG}
    j: PAlphaPartitionElement;
    f, c, na: TAlphaSet;
{$ENDIF}
begin
{$IFDEF DEBUG}
  // enumerate all elements (used for completeness check)
  f := TEmptyAlphaSet;
  j := FFront;
  while j <> nil do
  begin
    f := f + j^.alphaset;
    j := j^.next;
  end;
  f := f + AAlphaSet;
{$ENDIF}
  // @pre: Sets = { s_1, s_2, .., s_n } with
  //                  ( s_i * s_j ) = [] ForAll i,j with i<>j
  if AAlphaSet <> TEmptyAlphaSet then
  begin
    i := FFront;
    while i <> nil do
    begin
      // t <- intersection( current, input )
      t := ( i^.alphaset * AAlphaSet );
      if t <> TEmptyAlphaSet then
      begin
        // intersection non-empty -> fix for correct partitions
        if t = i^.alphaset then
        begin
          // Input is equal to, or superset of current
          // current is the subset
          // build @input={x| ( x in input ) and ( x not in current )}
          AAlphaSet := AAlphaSet - i^.alphaset;
        end else
        begin
          // Input and current have common elements
          if t = AAlphaSet then
          begin
            // Input is real subset of current,
            // build @current={x| ( x in current ) and ( x not in input )}
            i^.alphaset := i^.alphaset - AAlphaSet;
            // since current was a partion set and input part of this partition
            // there is no other partition with common elements -> stop, done
            break;
          end else
          begin
            // Input is neither superset nor subset of current
            // create 3 disjoint, non-empty sets:
            //   1. a = input-(input*current)
            //   2. b = current-(input*current)
            //   3. c = input*current
            // since: ( b union c ) = current, and current is a partition set,
            // b and c are partitions of this set too
            // -> remove current, add c,b and check the rest with a
            AAlphaSet := AAlphaSet - t; // a  -> rest
            i^.alphaset := i^.alphaset - t; // b -> delete current, add b
            frontAddSet( t ); // add c
          end;
        end;
      end;
      if AAlphaSet <> TEmptyAlphaSet then
        i := i^.next
      else
        i := nil;
    end;
    if AAlphaSet <> TEmptyAlphaSet then frontAddSet( AAlphaSet );
  end;
  // @post: @pre and
  //        Exists s' <= Sets with
  //          { e | ( e in s ) and ( s in s' ) } = @input
{$IFDEF DEBUG}
  // check partition criterion
  i := FFront;
  j := FFront;
  while i <> nil do
  begin
    j := i^.next;
    while j <> nil do
    begin
      assert( ( i^.alphaset * j^.alphaset ) = TEmptyAlphaSet );
      j := j^.next;
    end;
    i := i^.next;
  end;
  // check completness
  j := FFront;
  c := f;
  na := TAlphaSet([low(byte)..high(byte)]) - f;
  while j <> nil do
  begin
    c := c - j^.alphaset;
    na := na + j^.alphaset;
    j := j^.next;
  end;
  assert( c = TEmptyAlphaSet );
  assert( na = TAlphaSet([low(byte)..high(byte)]) );
{$ENDIF}
end;

function TAlphaSetPartition.Iterator: TAlphaSetPartitionIterator;
begin
  Result := TAlphaSetPartitionIterator.Create( FFront );
end;


{ TNodeSetIterator }

constructor TNodeSetIterator.Create(
  const ABVectorRef: TNodeSetField; ABSize: TFANodeType);
begin
  FBVectorRef := ABVectorRef;
  FSize := ABSize;
  Reset;
end;

procedure TNodeSetIterator.Reset;
begin
  FMask := TNodeSetFieldBlock( 0 );
  FAPos := -1;
  FCurrent := -1;
end;

function TNodeSetIterator.Next: Boolean;
begin
  if FCurrent < FSize then
  begin
    repeat
      Inc( FCurrent, 1 );
      FMask := TNodeSetFieldBlock( FMask ) shl TNodeSetFieldBlock( 1 );
      if FMask = TNodeSetFieldBlock( 0 ) then
      begin
        Inc( FAPos, 1 );
        FMask := TNodeSetFieldBlock( 1 );
      end;
    until ( FCurrent >= FSize ) or ( ( FBVectorRef[ FAPos ] and FMask ) <> TNodeSetFieldBlock( 0 ) );
  end;
  Result := FCurrent < FSize;
end;

function TNodeSetIterator.CurrentNode: TFANodeType;
begin
  Result := FCurrent;
end;

{ TNodeSet }

constructor TNodeSet.Create(ABSize: TFANodeType);
var i: Integer;
begin
  assert( ABSize >= 0 );
  FBSize := ABSize;
  SetLength( FBVector,
    ( ( ABSize + ( ( SizeOf( TNodeSetFieldBlock )*8 ) - 1 ) ) div
      ( SizeOf( TNodeSetFieldBlock )*8 ) ) );
  FHash := 0;
  for i := Low( FBVector ) to High( FBVector ) do
    FBVector[ i ] := 0;
end;

destructor TNodeSet.Destroy;
begin
  SetLength( FBVector, 0 );
  inherited Destroy;
end;

function TNodeSet.SetNode(n: TFANodeType; AIsIn: Boolean): Boolean;
var mask, bpos, oblock: TNodeSetFieldBlock;
    arrp: TFANodeType;
begin
  assert( n >= 0 );
  // ugly type enforcement ...
  arrp := n div TFANodeType( SizeOf( TNodeSetFieldBlock )*8 );
  bpos := n mod TFANodeType( SizeOf( TNodeSetFieldBlock )*8 );
  mask := ( TNodeSetFieldBlock( 1 ) shl TNodeSetFieldBlock( bpos ) );
  if AIsIn then
  begin
    oblock := FBVector[ arrp ];
    FBVector[ arrp ] := FBVector[ arrp ] or mask;
  end else
  begin
    oblock := FBVector[ arrp ];
    FBVector[ arrp ] := FBVector[ arrp ] and ( not mask );
  end;
  Result := oblock <> FBVector[ arrp ];
  if Result then FHash := FHash xor mask;
{$IFDEF DEBUG}
  if Result then
  begin
    arrp := n div TFANodeType( SizeOf( TNodeSetFieldBlock )*8 );
    bpos := n mod TFANodeType( SizeOf( TNodeSetFieldBlock )*8 );
    mask := ( TNodeSetFieldBlock( 1 ) shl TNodeSetFieldBlock( bpos ) );
    if AIsIn then
      Assert( ( FBVector[ arrp ] and mask ) > 0 )
    else
      Assert( ( FBVector[ arrp ] and mask ) = 0 );
  end;
{$ENDIF}
end;

function TNodeSet.IsEqual(ANodeSet: TNodeSet): Boolean;
var i: Integer;
begin
  Result := ( ANodeSet.FBSize = FBSize ) and ( ANodeSet.FHash = FHash );
  if Result then
  begin
    i := Low( FBVector );
    while Result and ( i <= High( FBVector ) ) do
    begin
      Result := FBVector[ i ] = ANodeSet.FBVector[ i ];
      Inc( i, 1 );
    end;
  end;
end;

function TNodeSet.Iterator: TNodeSetIterator;
begin
  Result := TNodeSetIterator.Create( FBVector, FBSize );
end;

{ TNodeSetMap }

function TNodeSetMap.mapNewNodeSet(const ANodeSet: TNodeSet): TFANodeType;
begin
  SetLength( FNodeSets, Succ( Length( FNodeSets ) ) );
  FNodeSets[ High( FNodeSets ) ] := ANodeSet;
  Result := High( FNodeSets );
end;

constructor TNodeSetMap.Create;
begin
  SetLength( FNodeSets, 0 );
end;

destructor TNodeSetMap.Destroy;
var i: Integer;
begin
  for i := Low( FNodeSets ) to High( FNodeSets ) do
    FNodeSets[ i ].Free;
  SetLength( FNodeSets, 0 );
  inherited Destroy;
end;

function TNodeSetMap.MapUnifyNodeSet(var ANodeSet: TNodeSet): TFANodeType;
var i: TFANodeType;
begin
  Result := TFANoNode;
  i := Low( FNodeSets );
  while ( Result = TFANoNode ) and ( i <= High( FNodeSets ) ) do
  begin
    if ( FNodeSets[ i ] = ANodeSet ) or
       ( FNodeSets[ i ].IsEqual( ANodeSet ) ) then
    begin
      Result := i;
      if FNodeSets[ i ] <> ANodeSet then
      begin
        ANodeSet.Free;
        ANodeSet := FNodeSets[ i ];
      end;
    end;
    Inc( i, 1 );
  end;
  if Result = TFANoNode then
    Result := mapNewNodeSet( ANodeSet );
end;

{ TDFAConstructor }

procedure TDFAConstructor.reInitialize;
begin
  if FNodeMap <> nil then FNodeMap.Free;
  if FDFA <> nil then FDFA.Free;
  if FAlpha <> nil then FAlpha.Free;
  if FTransSet <> nil then FTransSet.Free;
  FNodeMap := nil;
  FDFA := nil;
  FAlpha := nil;
  FTransSet := nil;
end;

function TDFAConstructor.getFinalState(
  const ANFANodeSet: TNodeSet): TFAFinalType;
begin
  case FMultiStateResolve of
    msrInputOrder: Result := getOrderFinalState( ANFANodeSet );
    msrException:
      begin
        Result := getNoMulFinalState( ANFANodeSet );
        if Result = TFAMultipleFinal then
          raise Exception.Create( 'DFA Node with Multiple Finalstates found' );
      end;
  end;
end;

function TDFAConstructor.getNoMulFinalState(
  const ANFANodeSet: TNodeSet): TFAFinalType;
var iter: TNodeSetIterator;
begin
  Result := TFANoFinal;
  iter := ANFANodeSet.Iterator;
  while ( Result <> TFAMultipleFinal ) and iter.Next do
  begin
    if FInputNFA.FNodeFinals[ iter.CurrentNode ] <> TFANoFinal then
    begin
      if Result = TFANoFinal then
      begin
        Result := FInputNFA.FNodeFinals[ iter.CurrentNode ];
      end else
      begin
        if Result <> FInputNFA.FNodeFinals[ iter.CurrentNode ] then
          Result := TFAMultipleFinal;
      end;
    end;
  end;
  iter.Free;
end;

function TDFAConstructor.getOrderFinalState(
  const ANFANodeSet: TNodeSet): TFAFinalType;
var iter: TNodeSetIterator;
begin
  Result := TFANoFinal;
  iter := ANFANodeSet.Iterator;
  while iter.Next do
  begin
    if ( FInputNFA.FNodeFinals[ iter.CurrentNode ] <> TFANoFinal ) and
       ( FInputNFA.FNodeFinals[ iter.CurrentNode ] >= Result ) then
    begin
      Result := FInputNFA.FNodeFinals[ iter.CurrentNode ];
    end;
  end;
  iter.Free;
end;

procedure TDFAConstructor.addDFANode(ANode: TFANodeType;
  AFinalState: TFAFinalType);
var tmpfin: String;
begin
  Assert( ANode = FDFA.getNodeCount );
  Assert( ANode >= TFAFinalMin );
  if AFinalState <> TFANoFinal then
    tmpfin := FInputNFA.FinalNames[ AFinalState ]
  else
    tmpfin := '';
  FDFA.addNode( TNFANode.Create, tmpfin );
end;

procedure TDFAConstructor.resolveEpsilons(var ANFANodeSet: TNodeSet);
var niter: TNodeSetIterator;
    titer: TNFATransitionIterator;
    done, changed: Boolean;
begin
  niter := ANFANodeSet.Iterator;
  done := false;
  while not done do
  begin
    done := true;
    niter.Reset;
    while niter.Next do
    begin
      titer := FInputNFA.Nodes[ niter.CurrentNode ].TransIterator;
      while titer.Next do
      begin
        if titer.CurrentTransSet = TEmptyAlphaSet then
        begin
          changed := ANFANodeSet.SetNode( titer.CurrentDestNode, true );
          done := done and ( not changed );
        end;
      end;
      titer.Free;
    end;
  end;
  niter.Free;
end;

function TDFAConstructor.getTransitionPartition(
  const ANFANodeSet: TNodeSet): TAlphaSetPartition;
var niter: TNodeSetIterator;
    titer: TNFATransitionIterator;
begin
  Result := TAlphaSetPartition.Create;
  niter := ANFANodeSet.Iterator;
  while niter.Next do
  begin
    titer := FInputNFA.Nodes[ niter.CurrentNode ].TransIterator;
    while titer.Next do
    begin
      Result.AddAlphaSet( titer.CurrentTransSet );
    end;
    titer.Free;
  end;
  niter.Free;
end;

function TDFAConstructor.resolveNode(const AStart: TNodeSet;
  const ATransitionSet: TAlphaSet): TFANodeType;
var target: TNodeSet;
    niter: TNodeSetIterator;
    titer: TNFATransitionIterator;
begin
  target := TNodeSet.Create( FInputNFA.NodeCount );
  // set bitflags for Nodes, which are reachable from @start
  // with Alphas @ATransitionSet (partitioned - this must be a subset)
  niter := AStart.Iterator;
  while niter.Next do
  begin
    titer := FInputNFA.Nodes[ niter.CurrentNode ].TransIterator;
    while titer.Next do
    begin
      if titer.CurrentTransSet <> TEmptyAlphaSet then
      begin
//        if ( ATransitionSet * titer.CurrentTransSet ) <> TEmptyAlphaSet then
        if ATransitionSet <= titer.CurrentTransSet then
        begin
          target.SetNode( titer.CurrentDestNode, true );

        end;
      end;
    end;
    titer.Free;
  end;
  niter.Free;
  // resolve Nodes reachable because of Epsilon Transitions
  resolveEpsilons( target );
  Result := FNodeMap.MapUnifyNodeSet( target );
end;

procedure TDFAConstructor.buildDFAInternals;
var cNode: TNodeSet;
    srcNodeID, destNodeID: TFANodeType;
    fstate: TFAFinalType;
begin
  // create startnode
  cNode := TNodeSet.Create( FInputNFA.getNodeCount );
  cNode.SetNode( FInputNFA.StartNode, true );
  resolveEpsilons( cNode );
  srcNodeID := FNodeMap.MapUnifyNodeSet( cNode );
  assert( srcNodeID = 0 );
  fstate := getFinalState( cNode );
  addDFANode( srcNodeID, fstate );
  FDFA.FStartNode := srcNodeID;
  // create follow nodes
  while srcNodeID < FDFA.getNodeCount do
  begin
    // create transition partition
    FTransSet := getTransitionPartition( FNodeMap.FNodeSets[ srcNodeID ] );
    // build target nodes
    FAlpha := FTransSet.Iterator;
    while FAlpha.Next do
    begin
      // build target nodeset which is reachable with the current AlphaPartition
      // map the nodeset
      destNodeID :=
        resolveNode( FNodeMap.FNodeSets[ srcNodeID ], FAlpha.CurrentSet );
      // check for new node -> add
      if destNodeID >= FDFA.getNodeCount then
      begin
        cNode := FNodeMap.FNodeSets[ destNodeID ];
        fstate := getFinalState( cNode );
        addDFANode( destNodeID, fstate );
      end;
      // add transition from srcNode to destNode with AlphaPartition
      FDFA.FNodes[ srcNodeID ].AddTransitionN( destNodeID, FAlpha.CurrentSet );
    end;
    FAlpha.Free;
    FTransSet.Free;
    FAlpha := nil;
    FTransSet := nil;
    // srcNode done with all transitions, set the next srcNode
    Inc( srcNodeID, 1 );
  end;
end;

constructor TDFAConstructor.Create;
begin
  FInputNFA := nil;
  FNodeMap := nil;
  FDFA := nil;
  FAlpha := nil;
  FTransSet := nil;
  FMultiStateResolve := msrInputOrder;
end;

destructor TDFAConstructor.Destroy;
begin
  reInitialize;
  inherited Destroy;
end;

function TDFAConstructor.ConstructDFA(const AInputNFA: TNFA): TNFA;
begin
  if AInputNFA.getNodeCount > 0 then
  begin
    FInputNFA := AInputNFA;
    FNodeMap := TNodeSetMap.Create;
    FDFA := TNFA.Create;
    buildDFAInternals;
    Result := FDFA;
    FDFA := nil;
  end else
  begin
    Result := TNFA.Create;
  end;
  Result.FDeterministic := true;
  reInitialize;
end;

{ TNodePartition }

constructor TNodePartition.Create(APartID: TPartitionID);
begin
  SetLength( FNodeRefs, 0 );
  FPartID := APartID;
end;

destructor TNodePartition.Destroy;
begin
  SetLength( FNodeRefs, 0 );
  inherited Destroy;
end;

procedure TNodePartition.AddNode(ANode: TFANodeType);
begin
  SetLength( FNodeRefs, Succ( Length( FNodeRefs ) ) );
  FNodeRefs[ High( FNodeRefs ) ] := ANode;
end;

{ TNodePartitionSet }

constructor TNodePartitionSet.Create;
begin
  SetLength( FParts, 0 );
end;

destructor TNodePartitionSet.Destroy;
var i: Integer;
begin
  for i := Low( FParts ) to High( FParts ) do
    FParts[ i ].Free;
  SetLength( FParts, 0 );
  inherited Destroy;
end;

procedure TNodePartitionSet.addPartition(APart: TNodePartition);
begin
  SetLength( FParts, Succ( Length( FParts ) ) );
  FParts[ High( FParts ) ] := APart;
end;

function TNodePartitionSet.GetPartition(APartID: TPartitionID): TNodePartition;
var i: Integer;
begin
  Result := nil;
  i := 0;
  while ( Result = nil ) and ( i <= High( FParts ) ) do
  begin
    if FParts[ i ].PartitionID = APartID then
      Result := FParts[ i ];
    Inc( i, 1 );
  end;
end;

procedure TNodePartitionSet.AddNode(APartID: TPartitionID; ANodeID: TFANodeType);
var part: TNodePartition;
begin
  part := getPartition( APartID );
  if part = nil then
  begin
    part := TNodePartition.Create( APartID );
    addPartition( part );
  end;
  part.AddNode( ANodeID );
end;

{ TNodePartitionMap }

function TNodePartitionMap.getPartCount: Integer;
begin
  Result := Length( FParts.Partitions );
end;

procedure TNodePartitionMap.remapNodesForPart(APart: TNodePartition);
var i: Integer;
begin
  for i := Low( APart.NodeRefs ) to High( APart.NodeRefs ) do
    FPartIDs[ APart.NodeRefs[ i ] ] := APart.PartitionID;
end;

constructor TNodePartitionMap.Create(ANodeNum: Integer);
var i: Integer;
    bpart: TNodePartition;
begin
  FTopPart := TPartitionIDMin;
  FParts := TNodePartitionSet.Create;
  SetLength( FPartIDs, ANodeNum );
  bpart := TNodePartition.Create( FTopPart );
  for i := Low( FPartIDs ) to High( FPartIDs ) do
  begin
    FPartIDs[ i ] := FTopPart;
    bpart.AddNode( i );
  end;
  FParts.addPartition( bpart );
end;

destructor TNodePartitionMap.Destroy;
begin
  FParts.Free;
  SetLength( FPartIDs, 0 );
  inherited Destroy;
end;

function TNodePartitionMap.PartIDLookup(ANodeId: TFANodeType): TPartitionID;
begin
  Result := FPartIDs[ ANodeId ];
end;

function TNodePartitionMap.GetPartition(APartID: TPartitionID): TNodePartition;
begin
  Result := FParts.GetPartition( APartID );
end;

function TNodePartitionMap.RefinePartition(ARefPartID: TPartitionID;
  var ASubPartition: TNodePartitionSet): Boolean;
var rpart: TNodePartition;
    i: Integer;
{$IFDEF DEBUG}
    j: Integer;
{$ENDIF}
begin
  // ! no deep checks, they are time consuming..
  // simply checking number of elements
  rpart := GetPartition( ARefPartID );
{$IFDEF DEBUG}
  j := 0;
  for i := Low( ASubPartition.Partitions ) to
    High( ASubPartition.Partitions ) do
    Inc( j, Length( ASubPartition.Partitions[ i ].NodeRefs ) );
  Assert( j = Length( rpart.NodeRefs ),
    'ASubPartition is not a refined Partition' );
{$ENDIF}
  if Length( ASubPartition.Partitions ) > 1 then
  begin
    Result := true;
    // hacky >_<
    // replace the reference partition with one of the new partitions
    i := 1;
    ASubPartition.Partitions[ 0 ].FPartID := ARefPartID;
    rpart.Free;
    FParts.FParts[ ARefPartID ] := ASubPartition.Partitions[ 0 ];
    remapNodesForPart( FParts.FParts[ ARefPartID ] );
    // create new partitions for what is left
    while i <= High( ASubPartition.Partitions ) do
    begin
      Inc( FTopPart, 1 );
      ASubPartition.Partitions[ i ].FPartID := FTopPart;
      FParts.addPartition( ASubPartition.Partitions[ i ] );
      remapNodesForPart( ASubPartition.Partitions[ i ] );
      Inc( i, 1 );
    end;
    SetLength( ASubPartition.FParts, 0 );
    ASubPartition.Free;
    ASubPartition := nil;
  end else Result := false; // else ignored.. partition with same size
end;

{ TMinDFAConstructor }

procedure TMinDFAConstructor.reInitialize;
begin
  if FPartMap <> nil then FPartMap.Free;
  if FMinDFA <> nil then FMinDFA.Free;
  FPartMap := nil;
  FMinDFA := nil;
end;

function TMinDFAConstructor.resolveTransSet(APart: TNodePartition): TAlphaSetPartition;
var i: Integer;
    titer: TNFATransitionIterator;
begin
  Result := TAlphaSetPartition.Create;
  for i := Low( APart.NodeRefs ) to High( APart.NodeRefs ) do
  begin
    titer := FInputNFA.Nodes[ APart.NodeRefs[ i ] ].TransIterator;
    while titer.Next do
    begin
      Result.AddAlphaSet( titer.CurrentTransSet );
    end;
    titer.Free;
  end;
end;

function TMinDFAConstructor.resolveDestPartSet(
  const AAlpha: TAlphaSet; APart: TNodePartition): TNodePartitionSet;
var i: Integer;
    titer: TNFATransitionIterator;
    done: Boolean;
begin
  Result := TNodePartitionSet.Create;
  for i := Low( APart.NodeRefs ) to High( APart.NodeRefs ) do
  begin
    titer := FInputNFA.Nodes[ APart.NodeRefs[ i ] ].TransIterator;
    done := false;
    // for each node, find the desstination node with given transition set
    while ( not done ) and titer.Next do
    begin
      if ( titer.CurrentTransSet * AAlpha ) <> TEmptyAlphaSet then
      begin
        Result.AddNode( FPartMap.PartIDLookup( titer.CurrentDestNode ),
                        APart.NodeRefs[ i ] );
        done := true;
      end;
    end;
    titer.Free;
    // if theres no such transition,
    // this transition would go into a (non existing)
    // error state -> assign pseudoid as partition id for the error node
    if not done then
      Result.AddNode( TPartitionPseudoID, APart.NodeRefs[ i ] );
  end;
end;

procedure TMinDFAConstructor.buildNodePartition;
var destset: TNodePartitionSet;
    transset: TAlphaSetPartition;
    titer: TAlphaSetPartitionIterator;
    i: Integer;
    done, alphadone: Boolean;
begin
  // setup first partitions -> partitionate into Non-Final and
  // Finals (each final gets another Partition since it should hit another
  // language)
  FPartMap := TNodePartitionMap.Create( FInputNFA.NodeCount );
  destset := TNodePartitionSet.Create;
  // ! TPartitionIDMin = 0, Finalstate can be < 0.. this works
  // because RefinePartition rewrites the Partition IDs!
  for i := Low( FInputNFA.Nodes ) to High( FInputNFA.Nodes ) do
    destset.AddNode( FInputNFA.FinalState[ i ], i );
  FPartMap.RefinePartition( FPartMap.FTopPart, destset );

  // refine partitions until nothing changes
  done := false;
  while not done do
  begin
    done := true;
    i := Low( FPartMap.FParts.FParts );
    while i <= High( FPartMap.FParts.FParts  ) do
    begin
      // take a partition with more than 1 node
      if Length( FPartMap.FParts.FParts[ i ].NodeRefs ) > 1 then
      begin
        // for all possible input elements, check
        // if transition creates a new partition
        transset := resolveTransSet( FPartMap.FParts.FParts[ i ] );
        titer := transset.Iterator;
        alphadone := true;
        while titer.Next and alphadone do
        begin
          destset := resolveDestPartSet( titer.CurrentSet,
                                         FPartMap.FParts.FParts[ i ] );
          if FPartMap.RefinePartition( FPartMap.FParts.FParts[ i ].PartitionID,
                                       destset ) then
            alphadone := false;
        end;
        titer.Free;
        done := done and alphadone;
      end;
      Inc( i, 1 );
    end;
  end;
end;

procedure TMinDFAConstructor.buildMinDFA;
var i, j: Integer;
    np: TNodePartition;
    cnode, dnode: TFANodeType;
    titer: TNFATransitionIterator;
    fin: TFAFinalType;
begin
  // construct new NFA
  FMinDFA := TNFA.Create;
  FMinDFA.FStartNode := TFANoNode;
  // Copy Finalstates, no changes here
  SetLength( FMinDFA.FFinals.FFinalStates, FInputNFA.FFinals.Count );
  for i := 0 to FInputNFA.FFinals.Count-1 do
    FMinDFA.FFinals.FFinalStates[ i ] := FInputNFA.FFinals.FFinalStates[ i ];
  // initialize Nodes and FinalState mapping of size Partitioncount
  SetLength( FMinDFA.FNodes, FPartMap.PartitionCount );
  SetLength( FMinDFA.FNodeFinals, FPartMap.PartitionCount );
  // Build new Nodes/Transitions/FinalMappings
  for i := TFANodeMin to High( FMinDFA.FNodes ) do
  begin
    fin := TFANoFinal;
    cnode := i;
    FMinDFA.FNodes[ cnode ] := TNFANode.Create;
    np := FPartMap.GetPartition( cnode );
    for j := Low( np.NodeRefs ) to High( np.NodeRefs ) do
    begin
      if np.NodeRefs[ j ] = FInputNFA.StartNode then
        FMinDFA.FStartNode := cnode;
      if fin = TFANoFinal then
        fin := FInputNFA.FinalState[ np.NodeRefs[ j ] ];
      titer := FInputNFA.Nodes[ np.NodeRefs[ j ] ].TransIterator;
      while titer.Next do
      begin
        dnode :=  FPartMap.PartIDLookup( titer.CurrentDestNode );
        FMinDFA.FNodes[ cnode ].AddTransitionN( dnode, titer.CurrentTransSet );
      end;
      titer.Free;
    end;
    FMinDFA.FNodeFinals[ cnode ] := fin;
  end;
end;

constructor TMinDFAConstructor.Create;
begin
  FInputNFA := nil;
  FMinDFA := nil;
  FPartMap := nil;
end;

destructor TMinDFAConstructor.Destroy;
begin
  reInitialize;
  inherited Destroy;
end;

function TMinDFAConstructor.ConstructMinDFA(const AInputNFA: TNFA): TNFA;
begin
  // this transformatuin doesnt work on "real" NFAs/ non-deterministic FAs
  if not AInputNFA.ConstructedDFA then
    raise TTransformationException.
      Create( 'NFA is flagged non-deterministic' );
  // No Finalstates -> partition algorithm doesnt work on this
  // ( the solutuion is a trival Non-Accepting automaton )
  if AInputNFA.FinalStateCount <= 0 then
    raise TTransformationException.
      Create( 'NFA without Final States not allowed' );

  FInputNFA := AInputNFA;
  buildNodePartition;
  buildMinDFA;
  Result := FMinDFA;
  FMinDFA := nil;
  Result.FDeterministic := true;
  Result.FMin := true;
  reInitialize;
end;

end.
