unit splex;

{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}
{$H+}

interface

uses Sysutils;

type
  TCharReaderStream = class
    public
      function NextChar: Boolean; virtual; abstract;
      function CurrentChar: Char; virtual; abstract;
      function StreamID: String; virtual; abstract;
  end;

  TTextReaderStream = class( TCharReaderStream )
    private
      FMatchString: String;
      FMatchLen: Integer;
      FCurrentState: Integer;
      FCurrentLine: Int64;
      FCurrentCol: Int64;
      FCurrentPos: Int64;
      FLNStyle: TTextLineBreakStyle;
    protected
      procedure SetStyle( ALNStyle: TTextLineBreakStyle );
      procedure Delta( const c: Char );
    public
      property CurrentCol: Int64 read FCurrentCol;
      property CurrentLine: Int64 read FCurrentLine;
      property CurrentPos: Int64 read FCurrentPos;
      property CurrentLNStyle: TTextLineBreakStyle read FLNStyle;
      constructor Create;
  end;

  TTextReaderFileStream = class( TTextReaderStream )
    private
      FFileName: String;
      FFileHandle: THandle;
      FOpen: Boolean;
      FBuffer: packed array of Char;
      FBufMax: Integer;
      FBufSize: Integer;
      FBufPos: Integer;
      function BufferNext: Boolean;
    public
      function NextChar: Boolean; override;
      function CurrentChar: Char; override;
      function StreamID: String; override;
      constructor Create( const AFileName: String; ABufMax: Integer = 1024 );
      destructor Destroy; override;
  end;

  { TTextReaderBufferStream }

  TTextReaderBufferStream = class( TTextReaderStream )
    private
      FBuffer: String;
      FBufPos: Integer;
    public
      function NextChar: Boolean; override;
      function CurrentChar: Char; override;
      function StreamID: String; override;
      constructor Create( const s: String );
  end;

type
  TTokenID = Integer;

const
  C_MaxTokenBuffer = 1024;

  // returned whenever the dfa stops in non-final state (which simply means
  // a character was read, at a wrong position or a given >enclosed< pattern
  // stopped without eos/a wrong character)
  L_NOSTATE = -1; 
  // End of Stream, returned at end of stream (or if there was nothing to read from the start )
  L_EOS = 0; 
  // splex generated Codes start at 1

type
  TSPLexPatternBuffer = array of Char;

  // scanner table structures
  TSPLexSigmaSet = set of Char;
  TSPLexSigmaTable = array of TSPLexSigmaSet;
  TSPLexSigmaRef = Integer;
  TSPLexNodeRef = Integer;
  TSPLexDeltaEntry = record
    sigma: TSPLexSigmaRef;
    dnode: TSPLexNodeRef;
  end;
  TSPLexDeltaTable = array of TSPLexDeltaEntry;
  TSPLexNodeEntry = record
    state: TTokenID;
    delta: TSPLexDeltaTable;
  end;
  TSPLexNodeTable = array of TSPLexNodeEntry;

  // scantable class
  TSPLexScanDFA = class
    private
      FCurrentNode: TSPLexNodeRef;
    protected
      FSigmaTable: TSPLexSigmaTable;
      FNodeTable: TSPLexNodeTable;
      FInitialNode: TSPLexNodeRef;
      function getCurrentState: TTokenID;
    public
      property State: TTokenID read getCurrentState;
      procedure Reset;
      function Delta( const s: Char ): Boolean;
      constructor Create; virtual;
      destructor Destroy; override;
  end;

  // generic scanner
  TSPLexScanner = class
    private
      FStreamOwner: Boolean;
      FTableOwner: Boolean;
      FInternalEOS: Boolean;
      FBuffered: Boolean;
      FMaxTokenBuffer: Integer;
      FInternalPattern: TSPLexPatternBuffer;
      FInternalPatternLength: Integer;
      FInternalToken: TTokenID;

      FPatternLine: Int64;
      FPatternCol: Int64;
      FPatternPos: Int64;
      function getLineStyle: TTextLineBreakStyle;
    protected
      FInternalStream: TTextReaderStream;
      FInternalDFA: TSPLexScanDFA;
      function NextChar: Boolean;
      function IsEOS: Boolean;
    public
      property StreamOwner: Boolean read FStreamOwner write FStreamOwner;
      property ScanTableOwner: Boolean read FTableOwner write FTableOwner;
      
      property MaxTokenBuffer: Integer read FMaxTokenBuffer;

      property LNStyle: TTextLineBreakStyle read getLineStyle;
      property StartCol: Int64 read FPatternCol;
      property StartLine: Int64 read FPatternLine;
      property StartPos: Int64 read FPatternPos;
      property CurrentToken: TTokenID read FInternalToken;
      property CurrentPattern: TSPLexPatternBuffer read FInternalPattern;
      property CurrentPatternLength: Integer read FInternalPatternLength;
      function CurrentStreamID: String;

      function NextToken: TTokenID; virtual;

      constructor Create( AScanStream: TTextReaderStream; AScanDFA: TSPLexScanDFA;
                          AMaxTokenBuffer: Integer = C_MaxTokenBuffer ); virtual;
      destructor Destroy; override;
  end;

implementation

constructor TTextReaderStream.Create;
begin
  SetLength( FMatchString, 0 );
  FMatchLen := 0;
  FCurrentState := 1;
  FCurrentLine := 1;
  FCurrentCol := 0;
  FCurrentPos := 0;
  if DefaultTextLineBreakStyle = tlbsLF then
    FLNStyle := tlbsCR
  else
    FLNStyle := tlbsLF;
  SetStyle( DefaultTextLineBreakStyle );
end;

procedure TTextReaderStream.SetStyle( ALNStyle: TTextLineBreakStyle );
begin
  if ALNStyle <> FLNStyle then
  begin
    case ALNStyle of
      tlbsLF: begin
                if ( FLNStyle = tlbsCRLF ) and
                   ( FCurrentState = 3 ) then
                  FCurrentState := 2
                else
                  FCurrentState := 1;
                FMatchLen := 1;
                FMatchString := #10;
              end;
      tlbsCR: begin
                if ( FLNStyle <> tlbsCRLF ) or
                   ( FCurrentState <> 2 ) then
                  FCurrentState := 1;
                FMatchLen := 1;
                FMatchString := #13;
              end;
      tlbsCRLF: begin
                if ( FLNStyle <> tlbsCR ) or
                   ( FCurrentState <> 2 ) then
                  FCurrentState := 1;
                FMatchLen := 2;
                FMatchString := String( #13 + #10 );
              end;
      else raise Exception.Create( 'Unknown TextLineBreakStyle' );
    end;
    FLNStyle := ALNStyle;
  end;
end;

procedure TTextReaderStream.Delta( const c: Char );
begin
  Inc( FCurrentPos, 1 );
  Inc( FCurrentCol, 1 );
  if FCurrentState > FMatchLen then
  begin
    Inc( FCurrentLine, 1 );
    FCurrentCol := 1;
    FCurrentState := 1;
  end;
  if FMatchString[ FCurrentState ] = c then
    Inc( FCurrentState, 1 );
end;

function TTextReaderFileStream.BufferNext: Boolean;
begin
  FBufSize := FileRead( FFileHandle, FBuffer[ 0 ], FBufMax );
  if FBufSize <= 0 then
  begin
    if FOpen then
    begin
      FileClose( FFileHandle );
      FOpen := false;
    end;
    Result := false;
  end else Result := true;
  FBufPos := 0;
end;

function TTextReaderFileStream.NextChar: Boolean;
begin
  if FOpen then
  begin
    Inc( FBufPos, 1 );
    Result := ( FBufPos < FBufSize ) or BufferNext;
    if Result then Delta( FBuffer[ FBufPos ] );
  end else Result := false;
end;

function TTextReaderFileStream.CurrentChar: Char;
begin
  Result := FBuffer[ FBufPos ];
end;

function TTextReaderFileStream.StreamID: String;
begin
  Result := FFileName;
end;

constructor TTextReaderFileStream.Create(const AFileName: String; ABufMax: Integer = 1024);
var i,j: Integer;
begin
  inherited Create;
  FFileName := AFileName;
  SetLength( FBuffer, ABufMax );
  FBufMax := ABufMax;
  FBufSize := 0;
  FBufPos := 0;
  if FileExists( AFileName ) then
  begin
    FFileHandle := FileOpen( AFileName, fmOpenRead );
    BufferNext;
    FOpen := FBufSize >= 0;
    FBufPos := -1;
  end else FOpen := false;
  if FOpen then
  begin
    // very simple linestyle detection
    // just assume the first "BufSize" characters contain newlines
    // the first newline style found is the one we use (otherwise just ignore
    // it.. users to be blamed )
    j := FBufSize-1;
    i := 0;
    while i < j do
    begin
      case FBuffer[ i ] of
        #10:
          begin 
            SetStyle( tlbsLF );
            i := j;
          end;
        #13:
          begin 
            if FBuffer[ i+1 ] = #10 then
            begin
              SetStyle( tlbsCRLF );
            end else
            begin
              SetStyle( tlbsCR );
            end;
            i := j;
          end;
      end;
      Inc( i, 1 );
    end;
  end;
end;

destructor TTextReaderFileStream.Destroy;
begin
  SetLength( FBuffer, 0 );
  if FOpen then FileClose( FFileHandle );
  inherited Destroy;
end;


function TSPLexScanner.getLineStyle: TTextLineBreakStyle;
begin
  Result := FInternalStream.CurrentLNStyle;
end;

function TSPLexScanner.NextToken: TTokenID;
var ndone: Boolean;
    cc: Char;
begin
  if not IsEOS then
  begin
    // set new pattern start
    FPatternLine := FInternalStream.CurrentLine;
    FPatternCol := FInternalStream.CurrentCol;
    FPatternPos := FInternalStream.CurrentPos;
    // reset DFA state ( startnode, initial state )
    FInternalDFA.Reset;
    // reset Pattern
    FInternalPatternLength := 0;
    // start reading characters and delta(state,char)->newstate with the dfa
    ndone := True;
    while ndone and NextChar do
    begin
      cc := FInternalStream.CurrentChar;
      ndone := FInternalDFA.Delta( cc );
      if ndone then
      begin
        // Consume Current "Symbol"
        FBuffered := False;
        // Append to Patternbuffer
        if FInternalPatternLength < FMaxTokenBuffer then
        begin
          FInternalPattern[ FInternalPatternLength ] := cc;
        end; // ignore
        Inc( FInternalPatternLength, 1 );
      end;
    end;
    Result := FInternalDFA.State;
    FInternalToken := Result;
    // -dont care about the buffer, when an error occured
    //  otherwise this could lead to an endless error loop on the same character
    // -dropping a character is an bad idea too, but it may be possible
    //  to recover and restart at another position (which is a bad idea too ;D)
    if Result = L_NOSTATE then FBuffered := false;
  end else
  begin
    FPatternLine := FInternalStream.CurrentLine;
    FPatternCol := FInternalStream.CurrentCol+1;
    FPatternPos := FInternalStream.CurrentPos+1;
    FInternalPatternLength := 0;
    FInternalToken := L_EOS;
    Result := L_EOS;
  end;
end;

function TSPLexScanner.CurrentStreamID: String;
begin
  Result := FInternalStream.StreamID;
end;

// Load LA Buffer if needed
function TSPLexScanner.NextChar: Boolean;
begin
  if not FBuffered then
  begin
    Result := FInternalStream.NextChar;
    if Result then
    begin
      FBuffered := True;
    end else FInternalEOS := true;
  end else Result := True;
end;

function TSPLexScanner.IsEOS: Boolean;
begin
  Result := FInternalEOS and ( not FBuffered );
end;

constructor TSPLexScanner.Create( AScanStream: TTextReaderStream;
                                  AScanDFA: TSPLexScanDFA;
                                  AMaxTokenBuffer: Integer = C_MaxTokenBuffer );
begin
  FStreamOwner := True;
  FTableOwner := True;
  FInternalEOS := False;
  FBuffered := False;
  FInternalStream := AScanStream;
  FInternalDFA := AScanDFA;
  FInternalDFA.Reset;
  FMaxTokenBuffer := AMaxTokenBuffer;
  SetLength( FInternalPattern, FMaxTokenBuffer );
  FInternalPatternLength := 0;
  FInternalToken := L_NOSTATE;
  // buffer first character and set position to first character (or 0 for eof)
  NextChar;
  FPatternLine := FInternalStream.CurrentLine;
  FPatternCol := FInternalStream.CurrentCol;
  FPatternPos := FInternalStream.CurrentPos;
end;

destructor TSPLexScanner.Destroy;
begin
  SetLength( FInternalPattern, 0 );
  if FStreamOwner and ( FInternalStream <> nil ) then FInternalStream.Free;
  if FTableOwner and ( FInternalDFA <> nil ) then FInternalDFA.Free;
end;

//

function TSPLexScanDFA.getCurrentState: TTokenID;
begin
  Assert( ( FCurrentNode >= Low( FNodeTable ) ) and
          ( FCurrentNode <= High( FNodeTable ) ) );

  Result := FNodeTable[ FCurrentNode ].state;
end;

procedure TSPLexScanDFA.Reset;
begin
  FCurrentNode := FInitialNode;
end;

function TSPLexScanDFA.Delta( const s: Char ): Boolean;
var i: Integer;
begin
  Assert( ( FCurrentNode >= Low( FNodeTable ) ) and
          ( FCurrentNode <= High( FNodeTable ) ) );

  Result := False;
  i := Low( FNodeTable[ FCurrentNode ].delta );
  while ( not Result ) and ( i <= High( FNodeTable[ FCurrentNode ].delta ) ) do
  begin
    if s in FSigmaTable[ FNodeTable[ FCurrentNode ].delta[ i ].sigma ] then
    begin
      Result := True;
      FCurrentNode := FNodeTable[ FCurrentNode ].delta[ i ].dnode;
    end;
    Inc( i, 1 );
  end;
end;

constructor TSPLexScanDFA.Create;
begin
  Reset;
end;

destructor TSPLexScanDFA.Destroy;
var i: TSPLexNodeRef;
begin
  SetLength( FSigmaTable, 0 );
  for i := Low( FNodeTable ) to High( FNodeTable ) do
    SetLength( FNodeTable[ i ].delta, 0 );
  SetLength( FNodeTable, 0 );
end;

{ TTextReaderBufferStream }

function TTextReaderBufferStream.NextChar: Boolean;
begin
  if FBufPos < Length(FBuffer) then
  begin
    Inc( FBufPos, 1 );
    Delta( FBuffer[ FBufPos ] );
    Result := true;
  end else Result := false;
end;

function TTextReaderBufferStream.CurrentChar: Char;
begin
  if FBufPos <= Length(FBuffer) then
    Result := FBuffer[ FBufPos ]
  else
    Result := #0;
end;

function TTextReaderBufferStream.StreamID: String;
begin
  Result := 'Generic_BufferStream';
end;

constructor TTextReaderBufferStream.Create(const s: String);
begin
  inherited Create;
  FBuffer := s;
  FBufPos := 0;
end;

end.
